# !/usr/bin/python

import logging
from typing import Any, Collection, Optional, Union

# noinspection PyUnresolvedReferences
from qgis.analysis import QgsGcpGeometryTransformer, QgsGcpTransformerInterface

# noinspection PyUnresolvedReferences
from qgis.core import (
    Qgis,
    QgsCoordinateReferenceSystem,
    QgsCoordinateTransform,
    QgsProject,
)

logger = logging.getLogger(__name__)

__all__ = ["transform_features", "transform_sub_tree_features"]


def transform_sub_tree_features(
    selected_nodes: Union[Any, Collection[Any]],
    transformer: QgsGcpGeometryTransformer,
    pre_transformer: Optional[Any] = None,
) -> None:
    # noinspection PyUnresolvedReferences
    from qgis.core import QgsLayerTreeGroup, QgsLayerTreeLayer, QgsLayerTreeNode

    if isinstance(selected_nodes, QgsLayerTreeLayer):
        transform_features(
            selected_nodes, transformer=transformer, pre_transformer=pre_transformer
        )
    elif isinstance(selected_nodes, QgsLayerTreeGroup):
        transform_sub_tree_features(
            selected_nodes.children(),
            transformer=transformer,
            pre_transformer=pre_transformer,
        )
    elif isinstance(selected_nodes, QgsLayerTreeNode):
        if selected_nodes.nodeType() == QgsLayerTreeNode.NodeGroup:
            transform_sub_tree_features(
                selected_nodes.children(),
                transformer=transformer,
                pre_transformer=pre_transformer,
            )
        else:
            logger.error(
                f"Node {selected_nodes} was not supported in transform_sub_tree_features, skipping"
            )
    else:
        if len(selected_nodes) == 0:
            logger.error(
                f"'Number of selected nodes was {len(selected_nodes)}, please supply some"
            )
            return

        for group in iter(selected_nodes):
            if isinstance(group, QgsLayerTreeLayer):
                transform_features(
                    group, transformer=transformer, pre_transformer=pre_transformer
                )
            elif isinstance(group, QgsLayerTreeGroup):
                transform_sub_tree_features(
                    group.children(),
                    transformer=transformer,
                    pre_transformer=pre_transformer,
                )
            elif isinstance(group, QgsLayerTreeNode):
                if group.nodeType() == QgsLayerTreeNode.NodeGroup:
                    transform_sub_tree_features(
                        group.children(),
                        transformer=transformer,
                        pre_transformer=pre_transformer,
                    )
                else:
                    logger.error(
                        f"Node {group} was not supported in transform_sub_tree_features, skipping"
                    )
            else:
                logger.error(
                    f"Node {group} was not supported in transform_sub_tree_features, skipping"
                )


def transform_features(
    tree_layer: Any,
    transformer: QgsGcpGeometryTransformer,
    pre_transformer: Optional[Any] = None,
) -> None:  #: QgsLayerTreeLayer
    """

    :param pre_transformer:
    :param transformer:
    :param tree_layer:
    :return:
    """

    if tree_layer is None:
        logger.error(f"Tree layer was None")
        return

    layer = tree_layer.layer()

    if not layer.isValid():
        logger.error(f"{layer.name()} is not valid!")
        return

    layer.startEditing()

    logger.warning(
        f"Transforming geometry of layer with name: {tree_layer.layer().name()}"
    )

    for idx, feat in enumerate(layer.getFeatures()):
        if not feat.hasGeometry():
            if False:
                assert (
                    feat.hasGeometry()
                ), f"Feature {idx} of {layer.name()} has no geometry"
            else:
                logger.error(
                    f"Feature {idx} of {layer.name()} has no geometry, skipping"
                )
                continue
        geometry = feat.geometry()
        if pre_transformer:
            geometry.transform(
                pre_transformer, Qgis.TransformDirection.ForwardTransform
            )

        geom, ok = transformer.transform(geometry)

        if pre_transformer:
            geom.transform(pre_transformer, Qgis.TransformDirection.ReverseTransform)

        if not ok:
            logger.error(
                f"Error while transforming {geom} in layer {tree_layer.layer().name()}"
            )
        feat.setGeometry(geom)
        layer.updateFeature(feat)

    layer.endEditCommand()
    layer.commitChanges()
    layer.triggerRepaint()



pipeline = """
{
    "pipeline":
        [
            {
                 "type": "readers.las",
                 "filename": "#"
            },
            {
                 "type": "filters.reprojection",
                 "in_srs": "EPSG:2193",
                 "out_srs": "EPSG:2193"
            },
            {
                 "type": "writers.las",
                 "filename": "#",
                 "a_srs": "EPSG:2193",
                 "forward": "all"
            }
        ]
}
"""

pdal_transformation_pipeline = """
[
    "untransformed.las",
    {
        "type":"filters.transformation",
        "matrix":"0 -1  0  1  1  0  0  2  0  0  1  3  0  0  0  1"
    },
    {
        "type":"writers.las",
        "filename":"transformed.las"
    }
]
"""

json = """
[
    "1.2-with-color.las",
    {
        "type": "filters.sort",
        "dimension": "X"
    }
]
"""

import pdal

pipeline = pdal.Pipeline(json)
count = pipeline.execute()
arrays = pipeline.arrays
metadata = pipeline.metadata
log = pipeline.log



r"""

from ezdxf.addons import odafc

# Load a DWG file
doc = odafc.readfile('my.dwg')

# Use loaded document like any other ezdxf document
print(f'Document loaded as DXF version: {doc.dxfversion}.')
msp = doc.modelspace()
...

# Export document as DWG file for AutoCAD R2018
odafc.export_dwg(doc, 'my_R2018.dwg', version='R2018')



ezdxf.addons.odafc.win_exec_path = "C:\Program Files\ODA\ODAFileConverter\ODAFileConverter.exe"


ezdxf.addons.odafc.unix_exec_path = "ODAFileConverter"

https://ezdxf.readthedocs.io/en/stable/addons/odafc.html
"""

from pathlib import Path
from typing import Optional

from ezdxf.addons import odafc

__all__ = ["convert_to_dxf"]


def convert_to_dxf(
    dwg_file_path: Path,
    oda_converter_path: Optional[
        Path
    ] = r"C:\Program Files\ODA\ODAFileConverter 25.4.0\ODAFileConverter.exe",
    target_dir: Optional[Path] = None,
) -> Path:
    # Load a DWG file
    # doc = odafc.readfile('my.dwg')

    # Use loaded document like any other ezdxf document
    # print(f'Document loaded as DXF version: {doc.dxfversion}.')
    # msp = doc.modelspace()
    # ...

    # Export document as DWG file for AutoCAD R2018
    # odafc.export_dwg(doc, 'my_R2018.dwg', version='R2018')

    if oda_converter_path:
        if not isinstance(oda_converter_path, Path):
            oda_converter_path = Path(oda_converter_path)

        odafc.win_exec_path = str(oda_converter_path)

    if not isinstance(dwg_file_path, Path):
        dwg_file_path = Path(dwg_file_path)

    if target_dir:
        if not isinstance(target_dir, Path):
            target_dir = Path(target_dir)

        if not target_dir.exists():
            target_dir.mkdir(parents=True, exist_ok=True)

        dxf_file_path = target_dir / dwg_file_path.name
    else:
        dxf_file_path = dwg_file_path.with_suffix(".dxf")

    odafc.convert(str(dwg_file_path), str(dxf_file_path), replace=True)

    return dxf_file_path


if __name__ == "__main__":

    def asdijai():
        win_path = Path(odafc.win_exec_path)
        print(win_path, win_path.exists())

        print(
            convert_to_dxf(
                Path(r"C:\Users\chen\Downloads\simple-christmas-objects.dwg")
            )
        )

    asdijai()



from collections import defaultdict
from copy import copy
from typing import Collection, Dict, List, Sequence, Union

import shapely
from shapely.constructive import simplify
from shapely.geometry.base import GeometrySequence
from tqdm import tqdm

from jord.geometric_analysis import construct_centerline
from jord.shapely_utilities import (
    clean_shape,
    dilate,
    extend_line,
    is_multi,
    iter_polygons,
    opening,
    pro_closing,
)
from jord.shapely_utilities.desliver_wkt import a_wkt
from jord.shapely_utilities.lines import (
    ExtensionDirectionEnum,
    find_isolated_endpoints,
    linemerge,
    snap_endings_to_points,
)
from jord.shapely_utilities.morphology import closing, erode, pro_opening
from jord.shapely_utilities.subdivision import subdivide

__all__ = ["desliver"]

from jord.shapely_utilities.projection import (
    get_min_max_projected_line,
    project_point_to_object,
)


def desliver(
    polygons: Collection[shapely.Polygon], buffer_size: float = 0.2
) -> List[shapely.geometry.Polygon]:
    buffered_exterior = []

    if isinstance(polygons, Sequence):
        polygons = list(polygons)

    for polygon in polygons:
        polygon: shapely.Polygon
        buffered_exterior.append(dilate(polygon, distance=buffer_size) - polygon)

    return buffered_exterior


def cut_polygon(
    polygon: shapely.Polygon, line_split_collection: List[shapely.LineString]
) -> GeometrySequence:
    line_split_collection.append(
        polygon.boundary
    )  # collection of individual linestrings for splitting in a list and add the polygon lines to it.
    merged_lines = linemerge(line_split_collection)
    border_lines = shapely.ops.unary_union(merged_lines)
    return shapely.ops.polygonize(border_lines)


def multi_line_extend(
    multi_line_string: Union[shapely.LineString, shapely.MultiLineString],
    distance: float,
) -> shapely.MultiLineString:
    isolated_endpoints = find_isolated_endpoints(multi_line_string)

    lines = []

    if isinstance(multi_line_string, shapely.LineString):
        ls = [multi_line_string]
    else:
        ls = multi_line_string.geoms

    for line in ls:
        start_point, end_point = shapely.Point(line.coords[0]), shapely.Point(
            line.coords[-1]
        )

        endpoint_in_isolated_points = end_point in isolated_endpoints

        direction = None
        if start_point in isolated_endpoints:
            if endpoint_in_isolated_points:
                direction = ExtensionDirectionEnum.both
            else:
                direction = ExtensionDirectionEnum.start
        elif endpoint_in_isolated_points:
            direction = ExtensionDirectionEnum.end

        if direction is not None:
            line = extend_line(line, offset=distance, simplify=False, side=direction)

        lines.append(line)

    return shapely.MultiLineString(lines)


def desliver_center_divide(
    polygons: Collection[shapely.Polygon],
    buffer_size: float = 0.2,
    post_process: bool = True,
    min_max_projection: bool = True,
    simplify_center_line: bool = False,
    close_res: bool = False,
) -> List[shapely.geometry.Polygon]:
    buffered_exterior = []

    if not isinstance(polygons, Sequence):
        polygons = list(polygons)

    for polygon in polygons:
        polygon: shapely.Polygon
        buffered_exterior.append(dilate(polygon, distance=buffer_size) - polygon)

    augmented_polygons = []

    intersections = []
    for ith in range(len(buffered_exterior)):
        a = buffered_exterior.copy()
        b = a.pop(ith)
        intersections.append(shapely.unary_union(a) & b)

    for ith, intersection in tqdm(enumerate(intersections)):
        minimum_clearance = intersection.minimum_clearance

        intersection = shapely.unary_union(list(iter_polygons(intersection)))

        center_line = construct_centerline(
            intersection, interpolation_distance=minimum_clearance / 3.14
        )

        if simplify_center_line:
            center_line = simplify_center_line(
                center_line, preserve_topology=False, tolerance=minimum_clearance / 2.0
            )

            center_line = simplify_center_line(
                center_line, preserve_topology=True, tolerance=minimum_clearance * 2.0
            )

        # TODO FIT LINE TO JAGGED LINE

        center_line = multi_line_extend(center_line, distance=minimum_clearance)

        if isinstance(intersection, shapely.Polygon):
            snapping_points = [
                shapely.Point(c) for c in subdivide(intersection).exterior.coords
            ]
        else:
            snapping_points = [
                shapely.Point(c)
                for inter in intersection.geoms
                for c in subdivide(inter).exterior.coords
            ]

        snapped_center_line = snap_endings_to_points(
            center_line, snapping_points=snapping_points, max_distance=minimum_clearance
        )

        poly = polygons[ith]

        if min_max_projection:
            for line in snapped_center_line.copy():
                projected_line = get_min_max_projected_line(line, poly)

                start, end = shapely.Point(projected_line.coords[0]), shapely.Point(
                    projected_line.coords[-1]
                )

                start_line, end_line = (
                    extend_line(
                        shapely.LineString(
                            (start, project_point_to_object(start, poly))
                        ),
                        offset=minimum_clearance,
                    ),
                    extend_line(
                        shapely.LineString((end, project_point_to_object(end, poly))),
                        offset=minimum_clearance,
                    ),
                )

                snapped_center_line.extend((start_line, end_line))

        res = cut_polygon(intersection, snapped_center_line)

        augmented = copy(poly)
        for r in res:
            un = r | poly
            re = erode(dilate(un, distance=1e-10), distance=1e-9)
            if is_multi(re):
                continue

            f = closing(un)

            if True:
                augmented |= f
            else:
                k = r & poly
                if k:
                    if isinstance(k, shapely.LineString):
                        if k.length >= minimum_clearance:
                            augmented |= r

        augmented = pro_opening(augmented, distance=minimum_clearance / 2.0)
        augmented_polygons.append(augmented)

    if post_process:
        if True:
            post_processed = []
            for ith in range(len(augmented_polygons)):
                a = augmented_polygons.copy()
                b = a.pop(ith)
                post_processed.append(
                    opening(
                        b - shapely.unary_union(a), distance=minimum_clearance / 2.0
                    ).simplify(
                        tolerance=minimum_clearance / 2.0, preserve_topology=False
                    )
                )
        else:
            post_processed = augmented_polygons

        if True:
            post_processed = shapely.MultiPolygon(post_processed).simplify(
                tolerance=minimum_clearance / 2.0, preserve_topology=True
            )

            post_processed_list = list(post_processed.geoms)
        else:
            post_processed_list = post_processed

        post_snaps = 2
        for _ in range(post_snaps):
            for ith in range(len(post_processed_list)):
                a = post_processed_list.copy()

                p = a.pop(ith)

                if True:  # Union
                    s = shapely.unary_union([clean_shape(l) for l in a])
                elif False:  # Deconstruct
                    if isinstance(s, shapely.geometry.Polygon):
                        coords = s.exterior.coords
                    else:
                        coords = []
                        for g in s.geoms:
                            coords.extend(g.exterior.coords)

                    s = [shapely.Point(c) for c in coords]
                else:
                    s = a

                if close_res:
                    ll = polygons.copy()
                    lp = ll.pop(ith)

                    post_processed_list[ith] = (
                        pro_closing(
                            shapely.snap(p, s, tolerance=minimum_clearance),
                            distance=minimum_clearance * 2.0,
                        )
                        | lp
                    ) - shapely.unary_union(ll)
                else:
                    post_processed_list[ith] = opening(
                        shapely.snap(p, s, tolerance=minimum_clearance),
                        distance=minimum_clearance * 2.0,
                    )

        return post_processed_list

    return augmented_polygons


def desliver_center_divide_shared(
    polygons: Collection[shapely.Polygon],
    buffer_size: float = 0.2,
    post_process: bool = True,
    min_max_projection: bool = False,
) -> List[shapely.geometry.Polygon]:
    buffered_exterior = []

    if isinstance(polygons, Sequence):
        polygons = list(polygons)

    for polygon in polygons:
        polygon: shapely.Polygon
        buffered_exterior.append(dilate(polygon, distance=buffer_size) - polygon)

    augmented_polygons = []

    intersections = []
    for ith in range(len(buffered_exterior)):
        a = buffered_exterior.copy()
        b = a.pop(ith)
        intersections.append(shapely.unary_union(a) & b)

    for ith, intersection in tqdm(enumerate(intersections)):
        minimum_clearance = intersection.minimum_clearance

        center_line = construct_centerline(
            intersection, interpolation_distance=minimum_clearance / 2.0
        )

        center_line = simplify(
            center_line, preserve_topology=False, tolerance=minimum_clearance / 2.0
        )

        center_line = simplify(
            center_line, preserve_topology=True, tolerance=minimum_clearance * 2.0
        )

        center_line = multi_line_extend(center_line, distance=minimum_clearance)

        if isinstance(intersection, shapely.Polygon):
            snapping_points = [
                shapely.Point(c) for c in subdivide(intersection).exterior.coords
            ]
        else:
            snapping_points = [
                shapely.Point(c)
                for inter in intersection.geoms
                for c in subdivide(inter).exterior.coords
            ]

        snapped_center_line = snap_endings_to_points(
            center_line, snapping_points=snapping_points, max_distance=minimum_clearance
        )

        poly = polygons[ith]

        if min_max_projection:
            for line in snapped_center_line.copy():
                projected_line = get_min_max_projected_line(line, poly)

                start, end = shapely.Point(projected_line.coords[0]), shapely.Point(
                    projected_line.coords[-1]
                )

                start_line, end_line = (
                    extend_line(
                        shapely.LineString(
                            (start, project_point_to_object(start, poly))
                        ),
                        offset=minimum_clearance,
                    ),
                    extend_line(
                        shapely.LineString((end, project_point_to_object(end, poly))),
                        offset=minimum_clearance,
                    ),
                )

                snapped_center_line.extend((start_line, end_line))

        res = cut_polygon(intersection, snapped_center_line)

        augmented = copy(poly)
        for r in res:
            un = r | poly
            re = erode(dilate(un, distance=1e-10), distance=1e-9)
            if is_multi(re):
                continue

            f = closing(un)

            if True:
                augmented |= f
            else:
                k = r & poly
                if k:
                    if isinstance(k, shapely.LineString):
                        if k.length >= minimum_clearance:
                            augmented |= r

        augmented = pro_opening(augmented, distance=minimum_clearance / 2.0)
        augmented_polygons.append(augmented)

    if post_process:
        post_processed = []

        for ith in range(len(augmented_polygons)):
            a = augmented_polygons.copy()
            b = a.pop(ith)
            post_processed.append(
                opening(b - shapely.unary_union(a), distance=minimum_clearance / 2.0)
            )

        # for p in post_processed:
        #  ...

        post_processed = list(
            shapely.MultiPolygon(post_processed)
            .simplify(tolerance=minimum_clearance / 2.0)
            .geoms
        )

        return post_processed

    return augmented_polygons


def desliver_least_intersectors_first(
    polygons: Collection[shapely.Polygon], buffer_size: float = 0.2
) -> Dict[int, shapely.geometry.Polygon]:
    buffered_exterior = []

    if isinstance(polygons, Sequence):
        polygons = list(polygons)

    for polygon in polygons:
        polygon: shapely.Polygon
        buffered_exterior.append(dilate(polygon, distance=buffer_size) - polygon)

    intersections = []
    for ith in range(len(buffered_exterior)):
        a = buffered_exterior.copy()
        b = a.pop(ith)
        intersections.append(shapely.unary_union(a) & b)

    inter_intersections = defaultdict(dict)
    num_intersections = len(intersections)
    for ith in range(num_intersections):
        for jth in range(num_intersections):
            if ith == jth:
                continue

            if (
                False
            ):  # TODO: OPTIMISATION when picking least intersectors to get intersection?
                if ith in inter_intersections[jth]:
                    continue

            c = intersections[ith] & intersections[jth]

            if not c.is_empty:
                inter_intersections[ith][jth] = c

    already_assigned = defaultdict(list)
    out = {}
    for ith_poly, intersectors in sorted(
        inter_intersections.items(), key=lambda d: len(d[-1].values()), reverse=False
    ):
        p = polygons[ith_poly]

        if intersectors:
            for ith_intersector, intersection in intersectors.items():
                already_assigned[ith_poly].append(ith_intersector)

                if ith_poly in already_assigned[ith_intersector]:
                    continue

                p |= intersection - polygons[ith_intersector]

            out[ith_poly] = pro_opening(p, distance=buffer_size)
        else:
            out[ith_poly] = p

    assert len(out) == len(polygons)

    return out


if __name__ == "__main__":

    def sauihd2():
        polygons = list(shapely.from_wkt(a_wkt).geoms)
        once = desliver_least_intersectors_first(polygons)
        out = desliver_least_intersectors_first(list(once.values()))

        c = shapely.MultiPolygon(iter_polygons(out.values()))
        ...

    def sauihd():
        polygons = list(shapely.from_wkt(a_wkt).geoms)
        once = desliver(polygons)
        out = desliver(list(once.values()))

        c = shapely.MultiPolygon(iter_polygons(out.values()))
        ...

    def sauihd3():
        polygons = list(shapely.from_wkt(a_wkt).geoms)
        once = desliver_center_divide(polygons)
        out = desliver_center_divide(list(once.values()))

        c = shapely.MultiPolygon(iter_polygons(out.values()))
        print(c.wkt)
        ...

    sauihd3()



import logging
from typing import Any, Mapping, Optional

from shapely import LineString, Point
from shapely.geometry.base import BaseGeometry

from .morphology import clean_shape, closing, opening, zero_buffer

__all__ = [
    "deflimmer",
    "clean_geometry",
    "unflimmer",
    "sanitise",
]

logger = logging.getLogger(__name__)
DEFAULT_DISTANCE = 1e-7


def deflimmer(geom: BaseGeometry, eps: float = DEFAULT_DISTANCE) -> BaseGeometry:
    """

    :param geom:
    :param eps:
    :return:
    """
    return opening(closing(geom, distance=eps), distance=eps)


clean_geometry = unflimmer = deflimmer


def sanitise(
    geom: BaseGeometry,
    *args: callable,
    kwargs: Optional[Mapping[callable, Mapping[str, Any]]] = None
) -> BaseGeometry:
    """
      #A positive distance produces a dilation, a negative distance an erosion. A very small or zero distance
      may sometimes be used to “tidy” a polygon.

    :param geom: The shape to sanitised
    :param args: The sanitisation callable steps
    :param kwargs: The sanitisation callable step kwargs in mappings with callable as key then sub-mapping is
    kwargs for callable
    :return: The sanitised shape
    """

    if kwargs is None:
        kwargs = {}

    if not len(args):
        args = (zero_buffer, deflimmer)

    for f in args:
        if f in kwargs:
            geom = f(geom, **(kwargs[f]))
        else:
            geom = f(geom)

    return geom


if __name__ == "__main__":

    def ausdhasu():
        p = Point((1, 1))
        print(clean_shape(p))

    def ausdhasu2():
        p = LineString([(1, 1), (1, 1)])
        print(clean_shape(p))

    ausdhasu2()



import numpy


def convert_raster_to_numpy_array(lyr) -> numpy.ndarray:
    """

    # Input: QgsRasterLayer

       :param lyr:
       :return:
       :rtype: numpy.ndarray
    """
    values = []
    provider = lyr.dataProvider()
    block = provider.block(1, lyr.extent(), lyr.width(), lyr.height())
    for i in range(lyr.height()):
        for j in range(lyr.width()):
            values.append(block.value(i, j))
    return numpy.array(values)



from typing import Union

import shapely
from shapely.geometry.linestring import LineString
from shapely.geometry.polygon import LinearRing, Polygon
from warg import pairs

from jord.shapely_utilities.lines import linemerge

__all__ = ["subdivide", "subdivide_polygon", "subdivide_line", "subdivide_ring"]


def subdivide(
    geom: Union[LineString, LinearRing, Polygon],
) -> Union[LineString, LinearRing, Polygon]:
    if isinstance(geom, LineString):
        return subdivide_line(geom)
    elif isinstance(geom, LinearRing):
        return subdivide_ring(geom)
    elif isinstance(geom, Polygon):
        return subdivide_polygon(geom)

    raise NotImplementedError(f"Subdivision for {type(geom)} not implemented")


def subdivide_line(line: LineString) -> LineString:
    half_point = line.interpolate(0.5, normalized=True)
    return shapely.LineString((line.coords[0], *half_point.coords, line.coords[-1]))


def subdivide_ring(ring: LinearRing) -> LinearRing:
    ring_segments = []
    for segment in pairs(list(ring.coords)):
        ring_segments.append(subdivide_line(LineString(segment)))

    return shapely.LinearRing(linemerge(ring_segments))


def subdivide_polygon(polygon: shapely.Polygon) -> shapely.Polygon:
    exterior = subdivide_ring(polygon.exterior)

    interiors = []
    for interior in polygon.interiors:
        interiors.append(subdivide_ring(interior))

    return Polygon(exterior, holes=interiors)


if __name__ == "__main__":

    def uihasud():
        from jord.shapely_utilities import dilate

        a = dilate(
            shapely.Point((0, 0)), distance=1, cap_style=shapely.BufferCapStyle.square
        )

        print(a.exterior)
        print(subdivide_ring(a.exterior))

    def uhasd():
        from jord.shapely_utilities import dilate

        a = dilate(
            shapely.Point((0, 0)), distance=1, cap_style=shapely.BufferCapStyle.square
        )
        b = dilate(
            shapely.Point((0, 0)), distance=0.5, cap_style=shapely.BufferCapStyle.square
        )
        c = a - b

        print(c)
        print(subdivide_polygon(c))

    uihasud()
    uhasd()



import random
from itertools import cycle
from typing import Callable, Generator, Iterable, Sized

from warg import QuadNumber, TripleNumber, n_uint_mix_generator_builder

__all__ = [
    "categorise_layer",
    "random_color_alpha_generator",
    "random_color_generator",
    "random_rgba",
    "random_rgb",
]


def random_rgb(mix: TripleNumber = (255, 255, 255)) -> TripleNumber:
    red = random.randrange(0, mix[0])
    green = random.randrange(0, mix[1])
    blue = random.randrange(0, mix[2])
    return red, green, blue


def random_rgba(mix: QuadNumber = (1, 1, 1, 1)) -> QuadNumber:
    red = random.randrange(0, mix[0])
    green = random.randrange(0, mix[1])
    blue = random.randrange(0, mix[2])
    alpha = random.randrange(0, mix[3])
    return red, green, blue, alpha


def random_color_generator() -> Generator[TripleNumber, None, None]:
    while 1:
        yield random_rgb()


def random_color_alpha_generator() -> Generator[QuadNumber, None, None]:
    while 1:
        yield random_rgba()


# noinspection PyUnresolvedReferences
from qgis.PyQt.QtGui import QColor

# noinspection PyUnresolvedReferences
from qgis.core import (
    QgsCategorizedSymbolRenderer,
    QgsRendererCategory,
    QgsSimpleFillSymbolLayer,
    QgsSymbol,
    QgsVectorLayer,
)

# noinspection PyUnresolvedReferences
from qgis.utils import iface


def categorise_layer(
    layer: QgsVectorLayer,
    field_name: str = "layer",
    *,
    color_iterable: Iterable = n_uint_mix_generator_builder(
        255, 255, 255, mix_min=(0, 0, 0)
    ),
    opacity: float = 1.0,
    outline_only: bool = False,
    outline_width=1.0,
) -> None:
    """

    https://qgis.org/pyqgis/3.0/core/Vector/QgsVectorLayer.html
    https://qgis.org/pyqgis/3.0/core/other/QgsFields.html

    :param layer:
    :param field_name:
    :param color_iterable:
    :return:
    """

    if isinstance(color_iterable, Sized):
        # noinspection PyTypeChecker
        color_iterable = cycle(color_iterable)

    if isinstance(color_iterable, Callable) and not isinstance(
        color_iterable, Generator
    ):
        # noinspection PyCallingNonCallable
        color_iterable = color_iterable()  # Compat

    color_iter = iter(color_iterable)

    available_field_names = layer.fields().names()

    assert (
        field_name in available_field_names
    ), f"Did not find {field_name=} in {available_field_names=}"

    render_categories = []
    for cat in layer.uniqueValues(layer.fields().indexFromName(field_name)):
        if cat is not None:
            sym = QgsSymbol.defaultSymbol(layer.geometryType())
            set_symbol_styling(color_iter, opacity, outline_only, outline_width, sym)

            render_categories.append(
                QgsRendererCategory(cat, symbol=sym, label=str(cat), render=True)
            )

    if True:  # add default
        sym = QgsSymbol.defaultSymbol(layer.geometryType())
        # https://qgis.org/pyqgis/master/core/QgsSymbolLayer.html#qgis.core.QgsSymbolLayer.setFillColor
        # https://qgis.org/pyqgis/3.40/core/QgsSimpleLineSymbolLayer.html
        # QgsLinePatternFillSymbolLayer

        # sym.symbolLayer(0).setStrokeColor(QColor(*col))
        # StrokeWidth
        # StrokeStyle
        # FillColor
        # FillStyle

        set_symbol_styling(color_iter, opacity, outline_only, outline_width, sym)

        render_categories.append(
            QgsRendererCategory("", symbol=sym, label="default", render=True)
        )

        if False:
            # render_categories.append(QgsRendererCategory()) # crashes qgis
            render_categories.append(
                QgsRendererCategory([], symbol=sym, label="EmptyList", render=True)
            )
            render_categories.append(
                QgsRendererCategory("", symbol=sym, label="EmptyString", render=True)
            )
            render_categories.append(
                QgsRendererCategory("None", symbol=sym, label="None", render=True)
            )

    layer.setRenderer(QgsCategorizedSymbolRenderer(field_name, render_categories))
    layer.triggerRepaint()
    iface.layerTreeView().refreshLayerSymbology(layer.id())


def set_symbol_styling(color_iter, opacity, outline_only, outline_width, sym) -> None:
    col = next(color_iter)
    if len(col) == 3:
        col = (*col, 255)
    if outline_only:
        outline_symbol_layer = QgsSimpleFillSymbolLayer()
        outline_symbol_layer.setColor(QColor("transparent"))
        outline_symbol_layer.setStrokeWidth(outline_width)
        outline_symbol_layer.setStrokeColor(QColor(*col))
        sym.changeSymbolLayer(0, outline_symbol_layer)
    else:
        sym.setColor(QColor(*col))
    sym.setOpacity(opacity)



from osgeo import osr

__all__ = ["get_spatial_reference_from_epsg"]


def get_spatial_reference_from_epsg(epsg_number: int) -> osr.SpatialReference:
    srs = osr.SpatialReference()
    res = srs.ImportFromEPSG(epsg_number)
    if res != 0:
        raise RuntimeError(f"{repr(res)}: could not import from EPSG")
    return srs



__author__ = "heider"
__doc__ = r"""

           Created on 1/27/23
           """

__all__ = []

import logging
from pathlib import Path

try:
    ...
except ImportError as ix:
    this_package_name = Path(__file__).parent.name
    logging.error(f"Make sure rasterio module is available for {this_package_name}")
    raise ix



__all__ = ["build_uri"]

import json
from typing import Any, Mapping, Optional


def build_uri(
    geom: Any,
    crs: Optional[str] = None,
    fields: Optional[Mapping[str, str]] = None,
    index: bool = False,
) -> str:
    """

    :param geom:
    :param crs:
    :param fields:
    :param index:
    :return:
    :rtype: str
    """

    geom_json = json.loads(geom.asJson())

    if geom_json is None:
        raise ValueError("Geom is empty")

    uri = geom_json["type"]  # As GeoJSON Repr, str dict

    if crs:
        uri += f"?crs={crs}"

    if fields:
        for k, v in fields.items():
            uri += f"&field={k}:{v}"

    uri += f'&index={"yes" if index else "no"}'

    return uri



from typing import List, Optional, Tuple

import shapely

__all__ = ["split_enveloping_geometry"]


def split_enveloping_geometry(
    geoms: List[shapely.geometry.base.BaseGeometry],
) -> Optional[
    Tuple[shapely.geometry.base.BaseGeometry, List[shapely.geometry.base.BaseGeometry]]
]:
    """
    Splits the enveloping geometry from the rest, if any, otherwise None.

    :param geoms:
    :return:
    """
    for geom in geoms:
        rest = set(geoms) - {geom}
        if all([shapely.contains(geom, r) for r in rest]):
            return geom, [*rest]



__all__ = ["load_spatialite"]

import sqlite3


def load_spatialite(dbapi_conn: sqlite3.Connection, *args, **kwargs) -> None:
    dbapi_conn.enable_load_extension(True)
    dbapi_conn.load_extension("mod_spatialite")



from typing import Iterable, Optional, Sequence, Tuple, Union

import numpy
import shapely
from shapely import LinearRing, MultiLineString
from shapely.geometry import LineString, Point, Polygon
from shapely.geometry.base import BaseGeometry
from warg import Number, pairs

from jord.shapely_utilities.morphology import dilate

__all__ = [
    "project_point_to_object",
    "project_point_to_line_points",
    "project_point_to_line",
    "nearest_geometry",
    "make_projected_ring",
    "make_extruded_ring",
    "get_min_max_projected_line",
    "line_line_intersection",
    "SingularExtentGeometry",
]

SingularExtentGeometry = Union[shapely.LineString, shapely.LinearRing, shapely.Polygon]


def project_point_to_object(point: Point, geometry: BaseGeometry) -> Point:
    """Find the nearest point in geometry, measured from given point.

    :param point: a shapely Point
    :param geometry: a shapely geometry object (LineString, Polygon)

    :return: a shapely Point that lies on geometry closest to point
    """
    nearest_point = None
    min_dist = float("inf")

    if isinstance(geometry, Polygon):
        for seg_start, seg_end in pairs(list(geometry.exterior.coords)):
            line_start = Point(seg_start)
            line_end = Point(seg_end)

            intersection_point = project_point_to_line_points(
                point, line_start, line_end
            )
            cur_dist = point.distance(intersection_point)

            if cur_dist < min_dist:
                min_dist = cur_dist
                nearest_point = intersection_point

    elif isinstance(geometry, LineString):
        for seg_start, seg_end in pairs(list(geometry.coords)):
            line_start = Point(seg_start)
            line_end = Point(seg_end)

            intersection_point = project_point_to_line_points(
                point, line_start, line_end
            )
            cur_dist = point.distance(intersection_point)

            if cur_dist < min_dist:
                min_dist = cur_dist
                nearest_point = intersection_point
    else:
        raise NotImplementedError(
            "project_point_to_object not implemented for"
            + " geometry type '"
            + geometry.type
            + "'."
        )
    return nearest_point


def project_point_to_line_points(
    point: Point, line_start: Point, line_end: Point, must_be_orthogonal: bool = False
) -> Point:
    """Find the nearest point on a straight line, measured from given point.

    Source: http://gis.stackexchange.com/a/438/19627

    :param must_be_orthogonal:
    :param point: a shapely Point object
    :param line_start: the line starting point as a shapely Point
    :param line_end: the line end point as a shapely Point

    :return: a shapely Point that lies on the straight line closest to point

    """

    line_magnitude = line_start.distance(line_end)

    u = (
        (point.x - line_start.x) * (line_end.x - line_start.x)
        + (point.y - line_start.y) * (line_end.y - line_start.y)
    ) / (line_magnitude**2)

    # closest point does not fall within the line segment,
    # take the shorter distance to an endpoint

    if u < 0.00001 or u > 1:
        ix = point.distance(line_start)
        iy = point.distance(line_end)

        if ix > iy:
            if must_be_orthogonal:
                raise Exception
            return line_end

        else:
            if must_be_orthogonal:
                raise Exception()
            return line_start

    ix = line_start.x + u * (line_end.x - line_start.x)
    iy = line_start.y + u * (line_end.y - line_start.y)

    return Point([ix, iy])


def project_point_to_line(point: Point, line: LineString) -> Point:
    line_coords = line.coords

    # assert line_coords == 2

    return project_point_to_line_points(point, *[Point(*xy) for xy in line_coords])


def line_line_intersection(line: LineString, other: LineString) -> Optional[Point]:
    """

    p = p1_start
    r = (p1_end - p1_start)

    q = p2_start
    s = (p2_end - p2_start)

    t = numpy.cross(q - p, s) / (numpy.cross(r, s))

    # This is the intersection point
    i = p + t * r

    :param line:
    :param other:
    :return:
    """

    import sympy.geometry

    l1 = sympy.geometry.Line(*[sympy.geometry.Point(*xy) for xy in line.coords])

    l2 = sympy.geometry.Line(*[sympy.geometry.Point(*xy) for xy in other.coords])

    l1_l2_intersection = l1.intersection(
        l2
    )  # These are two infinite lines defined by two points on the line

    if len(l1_l2_intersection) == 1:
        if isinstance(l1_l2_intersection[0], sympy.geometry.Line2D):  # Same
            return
        return Point(*l1_l2_intersection[0])


def get_intersection_linear_functions(
    a1: Number, b1: Number, a2: Number, b2: Number
) -> Tuple[Number, Number]:
    A = numpy.array([[-a1, 1], [-a2, 1]])
    b = numpy.array([[b1], [b2]])
    # you have to solve linear System AX = b where X = [x y]'
    return numpy.squeeze(numpy.linalg.pinv(A) @ b)  # x,y


def nearest_geometry(
    geometries: Sequence[BaseGeometry], point: Point
) -> Tuple[BaseGeometry, float, int]:
    """Find the nearest geometry among a list, measured from fixed point.

    :param geometries: a list of shapely geometry objects
    :param point: a shapely Point

    :return:        Tuple (geom, min_dist, min_index) of the geometry with minimum distance        to point,
    its distance min_dist and the list index of geom, so that        geom = geometries[min_index].
    """
    min_dist, min_index = min(
        (point.distance(geom), k) for (k, geom) in enumerate(geometries)
    )

    return geometries[min_index], min_dist, min_index


def get_min_max_projected_line(
    geom: SingularExtentGeometry, other: shapely.geometry.base.BaseGeometry
) -> shapely.LineString:
    if not isinstance(geom, shapely.LineString):
        geom = geom.boundary

    if isinstance(other, shapely.MultiLineString):
        geom = geom.boundary

    min_v = max_v = 0.5

    other_coords = other.boundary.coords

    # Find limits
    for point_coords in other_coords:
        v = geom.project(Point(point_coords), normalized=True)
        if v < min_v:
            min_v = v
        elif v > max_v:
            max_v = v
        else:
            ...

    # Reconstruct line within limit
    vs = []
    for point_coords in geom.coords:
        p = Point(point_coords)
        d = geom.project(p, normalized=True)
        if max_v > d > min_v:
            vs.append(d)
        else:
            ...

    # translate distances to points
    coords = []
    for d in [min_v, *sorted(vs), max_v]:
        a = geom.interpolate(d, normalized=True)
        coords.append(a)

    return LineString(coords)


def make_projected_ring(
    lines: Union[MultiLineString, Iterable[LineString]], ccw: bool = True
) -> LinearRing:
    points = []

    if isinstance(lines, MultiLineString):
        lines = lines.geoms

    num_lines = len(lines)

    if ccw:
        lines = lines[::-1]

    for n in range(num_lines):
        points.append(project_point_to_line(Point(lines[n - 1].coords[-1]), lines[n]))

    ring = LinearRing(points)

    assert ring.is_closed
    assert ring.is_ring

    return ring


def make_extruded_ring(
    lines: Union[MultiLineString, Iterable[LineString]], ccw: bool = True
) -> LinearRing:
    points = []

    if isinstance(lines, MultiLineString):
        lines = lines.geoms

    num_lines = len(lines)

    if ccw:
        lines = lines[::-1]

    for n in range(num_lines):
        points.append(Point(line_line_intersection(lines[n - 1], lines[n]).coords[-1]))

    ring = LinearRing(points)

    assert ring.is_closed
    assert ring.is_ring

    return ring


if __name__ == "__main__":

    def uihasuih():
        print(
            line_line_intersection(
                LineString([[0, 0], [1, 1]]), LineString([[1, 0], [0, 1]])
            )
        )
        print(
            line_line_intersection(
                LineString([[0, 0], [1, 1]]), LineString([[6, 0], [5, 1]])
            )
        )
        print(
            line_line_intersection(
                LineString([[0, 0], [1, 1]]), LineString([[0, 0], [1, 1]])
            )
        )
        print(
            line_line_intersection(
                LineString([[0, 0], [1, 1]]), LineString([[1, 0], [2, 1]])
            )
        )

    def uijhas():
        line = LineString([[0, 0], [1, 1], [2, 0], [1, -1]])
        poly = dilate(Point((1, 0)), distance=0.4)
        print(line)
        print(poly)
        print(get_min_max_projected_line(line, poly).wkt)

    def juijh():
        r"""

            0   1   2

        0   0---0   0
                    |
        1   0       0
            |
        2   0   0---0

        to become


            0   1   2

        0   0---0---0
            |       |
        1   0       0
            |       |
        2   0---0---0

        :return:
        """

        lines = [
            LineString([[0, 0], [1, 0]]),
            LineString([[2, 0], [2, 1]]),
            LineString([[2, 2], [1, 2]]),
            LineString([[0, 2], [0, 1]]),
        ]

        print(make_extruded_ring(lines))

    def juijh2():
        r"""

            0   1   2   3

        0   0---0---0---0
                    |
        1   0       0
            |
        2   0   0---0

        to become


            0   1   2

        0   0---0---0
            |       |
        1   0       0
            |       |
        2   0---0---0

        :return:
        """

        lines = [
            LineString([[0, 0], [3, 0]]),
            LineString([[2, 0], [2, 1]]),
            LineString([[2, 2], [1, 2]]),
            LineString([[0, 2], [0, 1]]),
        ]

        print(make_extruded_ring(lines))

    # juijh2()

    uijhas()



from qgis.core import QgsEditorWidgetSetup, QgsVectorLayer

__all__ = ["add_value_relation_field"]


def add_value_relation_field(
    layer: QgsVectorLayer,
    field_name: str,
    referenced_layer: QgsVectorLayer,
    key_field: str,
    value_field: str,
    allow_null: bool = True,
    allow_multi: bool = False,
    filter_expression: str = "",
) -> None:
    """
    Adds a value relation widget to a field in the layer.

    Args:
        layer: The layer containing the field to set up
        field_name: Name of the field to set up
        referenced_layer: Layer containing the values to choose from
        key_field: Field in referenced layer to use as key
        value_field: Field in referenced layer to display to user
        allow_null: Allow NULL values
        allow_multi: Allow multiple value selection
        filter_expression: Optional filter expression for referenced layer
    """
    config = {
        "AllowMulti": allow_multi,
        "AllowNull": allow_null,
        "FilterExpression": filter_expression,
        "Key": key_field,
        "Layer": referenced_layer.id(),
        "NofColumns": 1,
        "OrderByValue": False,
        "UseCompleter": False,
        "Value": value_field,
    }

    setup = QgsEditorWidgetSetup("ValueRelation", config)
    layer.setEditorWidgetSetup(layer.fields().indexOf(field_name), setup)


def highlight_relationship():
    from qgis.utils import iface
    from qgis.core import QgsProject, QgsFeatureRequest

    parent = iface.activeLayer()  # Get parent layer from the ToC

    def selectChildren(fids, foo, bar):
        relations = QgsProject.instance().relationManager().relationsByName("my_rel")
        rel = relations[0]  # Assuming we only have 1 relation named 'my_rel'
        referencingLayer = rel.referencingLayer()
        referencedLayer = rel.referencedLayer()

        request = QgsFeatureRequest().setFilterFids(fids)
        fit = referencedLayer.getFeatures(request)
        childIds = []
        for f in fit:
            it = rel.getRelatedFeatures(f)
            childIds.extend([i.id() for i in it])

        iface.mapCanvas().flashFeatureIds(referencingLayer, childIds)

    parent.selectionChanged.connect(selectChildren)


if __name__ == "__main__":
    ...
    # add_value_relation_field(      layer=my_layer,      field_name='category',
    # referenced_layer=categories_layer,      key_field='id',      value_field='name'      )



# !/usr/bin/env python3

from typing import Iterable

# noinspection PyUnresolvedReferences
from qgis.PyQt.QtGui import QColor

# noinspection PyUnresolvedReferences
from qgis.core import (
    QgsCategorizedSymbolRenderer,
    QgsLineSymbol,
    QgsRendererCategory,
    QgsSymbol,
    QgsVectorLayer,
)

__all__ = ["set_geometry_constraints"]


def set_geometry_constraints(layers: QgsVectorLayer) -> None:
    if layers is None:
        return

    geometry_options = [
        "QgsIsValidCheck",
    ]

    gap_check_configuration = {
        "allowedGapsBuffer": 0,
        "allowedGapsEnabled": False,
        "allowedGapsLayer": "",
    }

    for layers_inner in layers:
        if layers_inner:
            if isinstance(layers_inner, Iterable):
                for layer in layers_inner:
                    if layer:
                        layer.geometryOptions().setGeometryChecks(geometry_options)
                        # layer.geometryOptions().setCheckConfiguration("QgsGeometryGapCheck", gap_check_configuration)
                        # layer.geometryOptions().setGeometryPrecision(0.0001)
                        layer.geometryOptions().setRemoveDuplicateNodes(True)
            else:
                layers_inner.geometryOptions().setGeometryChecks(geometry_options)
                layers_inner.geometryOptions().setRemoveDuplicateNodes(True)



__all__ = ["plugin_status"]


def plugin_status(plugin_key: str) -> str:
    # noinspection PyUnresolvedReferences
    from pyplugin_installer.installer_data import repositories, plugins  # QGIS CODE!

    all_plugins = plugins.all()

    if plugin_key in all_plugins:
        return all_plugins[plugin_key]["status"]

    return "Not Found"


# utils.reloadPlugin()



from geopandas import GeoDataFrame


def raster_window(
    window_min_x,
    window_min_y,
    window_size_x,
    window_size_y,
    image_resolution_x,
    image_resolution_y,
    features,
    window_features,
):
    window_df = GeoDataFrame({"geometry": window_features})

    # Creates upscaled resolution to initially create image
    # Then it is scaled down
    # This is done for higher quality output image
    upscaled_resolution_x = image_resolution_x * 4
    upscaled_resolution_y = image_resolution_y * 4

    # Scales up geometries to fit the image resolution and translates the geometries to have minimum at (0,0)
    x_scale = upscaled_resolution_x / window_size_x
    y_scale = upscaled_resolution_x / window_size_x
    x_offset = -window_min_x * upscaled_resolution_x / window_size_x
    y_offset = -window_min_y * upscaled_resolution_y / window_size_y

    affine_transformation_matrix = [x_scale, 0, 0, y_scale, x_offset, y_offset]
    window_df["transformed_geometry"] = window_df["geometry"].affine_transform(
        affine_transformation_matrix
    )

    rasterized = features.rasterize(
        window_df["transformed_geometry"].tolist(),
        out_shape=(upscaled_resolution_x, upscaled_resolution_y),
        fill=1,
        out=None,
        all_touched=True,
        default_value=0,
        dtype=None,
    )

    return rasterized



from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Type, Union

import shapely
from osgeo import ogr

from jord.gdal_utilities.spatial_reference import get_spatial_reference_from_epsg

__all__ = [
    "create_ogr_table",
    "insert_ogr_table",
    "ShapelyGeometry",
    "FIELDS_TYPES_MAP",
    "OgrGeometryTypeEnum",
]


class OgrGeometryTypeEnum(Enum):
    point = ogr.wkbPoint
    line_string = ogr.wkbLineString
    polygon = ogr.wkbPolygon


FIELDS_TYPES_MAP = {  # https://gis.stackexchange.com/questions/18715/mapping-between-ogr-and-python-data-types
    int: ogr.OFTInteger,  # Simple 32bit integer
    float: ogr.OFTReal,  # Double Precision floating point
    str: ogr.OFTString,  # String of ASCII chars
}


@dataclass
class ShapelyGeometry:
    geometry: shapely.geometry.base.BaseGeometry


def create_ogr_table(
    spatialite_db: Path,
    table_name: str,
    fields: list[tuple[str, Type[Union[int, float, str]]]],
    srs_id: int,
) -> None:
    """

    :param spatialite_db: The spatialite_db to create table in
    :param table_name:
    :param fields: A list of fields to create in the table. Each field is a tuple defined as ("column_name",
    Type)
    :param srs_id: Spatial reference system id of the geometries.
    :return:
    """
    sqlite_dr = ogr.GetDriverByName("SQLite")
    sl_ds = sqlite_dr.Open(str(spatialite_db), update=1)
    sl_lyr = sl_ds.CreateLayer(
        table_name,
        srs=get_spatial_reference_from_epsg(srs_id),
        geom_type=OgrGeometryTypeEnum.polygon.value,
    )

    for field in fields:
        sl_lyr.CreateField(ogr.FieldDefn(field[0], FIELDS_TYPES_MAP[field[1]]))

    del sl_ds  # clean up


def insert_ogr_table(
    spatialite_db: Path, geoms: list[ShapelyGeometry], table_name: str
) -> None:
    """
    Inserts a list of ShapelyGeometry into table.

    :param spatialite_db:
    :param geoms:
    :param table_name:
    :return:
    """
    sl_ds = ogr.Open(str(spatialite_db), update=1)
    sl_lyr = sl_ds.GetLayerByName(table_name)

    feature_def = sl_lyr.GetLayerDefn()

    sl_lyr.StartTransaction()

    for g in geoms:
        geom = ogr.CreateGeometryFromWkb(g.geometry.wkb)
        geom.FlattenTo2D()

        dst_feat = ogr.Feature(feature_def)
        dst_feat.SetGeometryDirectly(geom)
        fields = g.__dict__
        fields.pop("geometry")
        for (
            field,
            value,
        ) in (
            fields.items()
        ):  # Add all fields - 'geometry' in SpacePolygon to the feature
            dst_feat.SetField(field, value)

        sl_lyr.CreateFeature(dst_feat)
        dst_feat = geom = None  # destroy these

    sl_lyr.CommitTransaction()
    del sl_ds



from typing import Dict

from jord.gdal_utilities import OGR

__all__ = ["available_driver_index_name_mapping"]


def available_driver_index_name_mapping() -> Dict[str, int]:
    driver_map = {}
    for i in range(OGR.GetDriverCount()):
        driver_name = OGR.GetDriver(i).GetName()
        assert driver_name not in driver_map
        driver_map[driver_name] = i

    return driver_map


if __name__ == "__main__":
    print(available_driver_index_name_mapping())



from typing import Dict

from geopandas import GeoDataFrame

from jord.shapely_utilities import ShapelyGeometryTypesEnum

__all__ = ["split_on_geom_type"]


def split_on_geom_type(
    data_frame: GeoDataFrame,
) -> Dict[ShapelyGeometryTypesEnum, GeoDataFrame]:
    """options for filtering geometries in GeoDataFrame's, rather messy interface tbh, just like pandas.

    gdf0 = gdf.loc[gdf.geometry.geometry.type=='MultiPolygon']

    gdf1 = gdf[gdf.geometry.apply(lambda x : x.type=='MultiPolygon')]

    gdf2 = gdf[gdf.geom_type=='MultiPolygon']

    gdf3 = gdf[gdf.geometry.type=="MultiPolygon"]
    """
    return {
        t: data_frame[data_frame.geom_type == t.value.__name__]
        for t in ShapelyGeometryTypesEnum
    }



__author__ = "heider"
__doc__ = r"""

           Created on 5/5/22
           """

from PIL.TiffTags import TAGS, TAGS_V2
from warg import invert_dict, pivot_dict_object

__all__ = ["TIFF_TAG_IDS", "TIFF_TAG_V2_IDS"]

TIFF_TAG_IDS = invert_dict(TAGS)

# TIFF_TAG_V2_IDS = invert_dict(TAGS_V2) # NOT hashable value types
TIFF_TAG_V2_IDS = pivot_dict_object(TAGS_V2, "name")

if __name__ == "__main__":
    print(TIFF_TAG_IDS)
    print(TIFF_TAG_V2_IDS)



from PyQt5.QtGui import QTransform

__all__ = ["read_wld_file"]


def read_wld_file(geom, wld_file_path):
    assert wld_file_path is not None
    assert wld_file_path.exists()
    with open(wld_file_path) as wld_file:
        m32 = (float(c) for c in wld_file.readlines())

        transformer = QTransform(*m32)

        geom.transform(transformer)



import logging
from typing import Any, Iterable, Mapping, Optional

# noinspection PyUnresolvedReferences
from qgis.PyQt.QtGui import QColor

# noinspection PyUnresolvedReferences
from qgis.core import (
    QgsCategorizedSymbolRenderer,
    QgsLineSymbol,
    QgsPalLayerSettings,
    QgsRendererCategory,
    QgsSymbol,
    QgsTextBackgroundSettings,
    QgsTextFormat,
    QgsVectorLayer,
    QgsVectorLayerSimpleLabeling,
)
from warg import TripleNumber

__all__ = [
    "style_layer_from_mapping",
    "set_3d_view_settings",
    "set_label_styling",
    "set_layer_rendering_scale",
]

from jord.qgis_utilities.enums import (
    Qgis3dAltitudeClamping,
    Qgis3dAltitudeBinding,
    Qgis3dCullingMode,
    Qgis3dFacade,
)

logger = logging.getLogger(__name__)


def style_layer_from_mapping(
    layer: QgsVectorLayer,
    style_mapping_field_dict: Mapping,
    field_name: str = "layer",
    default_color: TripleNumber = (0, 0, 0),
    default_opacity: float = 1.0,
    default_width: float = 0.1,
    *,
    repaint: bool = True,
) -> None:
    if layer is None:
        return

    style_mapping = style_mapping_field_dict[field_name]

    render_categories = []
    for cat in layer.uniqueValues(layer.fields().indexFromName(field_name)):
        cat_color = default_color
        cat_opacity = default_opacity
        cat_width = default_width
        label = str(cat)

        if cat in style_mapping.keys():
            style = style_mapping[label]
            if "color" in style:
                cat_color = (
                    int(n) for n in style["color"]
                )  # TODO: also support with AlphaChannel | Qt.GlobalColor | QGradient
            if "opacity" in style:
                cat_opacity = max(0.0, min(float(style["opacity"]), 1.0))
            if "width" in style:
                cat_width = max(0.0, float(style["width"]))

        symbol = QgsSymbol.defaultSymbol(layer.geometryType())
        symbol.setColor(QColor(*cat_color, 255))
        symbol.setOpacity(cat_opacity)

        if isinstance(symbol, QgsLineSymbol):
            symbol.setWidth(cat_width)
        else:
            logging(f"width ignored, symbol is of type: {type(symbol)}")

        render_categories.append(
            QgsRendererCategory(cat, symbol=symbol, label=label, render=True)
        )

    layer.setRenderer(QgsCategorizedSymbolRenderer(field_name, render_categories))
    if repaint:
        layer.triggerRepaint()


def set_layer_rendering_scale(
    layers: QgsVectorLayer,
    *,
    max_ratio: float = 1.0,
    min_ratio: float = 1 / 9999,
):
    """

    :param layers:
    :param max_ratio:
    :param min_ratio:
    :return:
    """
    if not isinstance(
        layers, Iterable
    ):  # Handle both single layer and iterable of layers
        layers = [layers]

    for layer in layers:
        if not layer:
            continue

        layer.setScaleBasedVisibility(True)
        layer.setMaximumScale(max_ratio)
        layer.setMinimumScale(1 / min_ratio)


def set_label_styling(
    layers: QgsVectorLayer,
    *,
    field_name: str = "name",
    max_ratio: float = 1.0,
    min_ratio: float = 1 / 999,
    font_size: int = 10,
    background: bool = False,
    background_color: tuple = (255, 255, 255, 200),
    background_svg: str = None,
    html_format: Optional[str] = None,
):
    if layers is None:
        return

    # Handle both single layer and iterable of layers
    if not isinstance(layers, Iterable):
        layers = [layers]

    for layer in layers:
        if not layer:
            continue

        # Create label settings
        label_settings = QgsPalLayerSettings()

        # Set field name with optional HTML formatting
        if html_format:  # ACTUALLY USE HTML
            # label_settings.isExpression = True
            label_settings.allowHtml = True
            label_settings.fieldName = (
                f"'<div style=\"text-align: center;\">' || {field_name} || '</div>'"
            )
        else:
            label_settings.fieldName = field_name

        # Set font settings
        format = QgsTextFormat()
        format.setSize(font_size)
        format.setNamedStyle("Regular")
        format.setColor(QColor(0, 0, 0))

        # Set background if enabled
        if background:
            from qgis.PyQt.QtCore import QSizeF

            bg_buffer = QgsTextBackgroundSettings()
            bg_buffer.setEnabled(True)
            bg_buffer.setFillColor(QColor(*background_color))
            bg_buffer.setSize(QSizeF(1, 0.5))
            bg_buffer.setType(QgsTextBackgroundSettings.ShapeRectangle)

            if background_svg:
                bg_buffer.setSvgFile(background_svg)
                bg_buffer.setType(QgsTextBackgroundSettings.ShapeSVG)

            format.setBackground(bg_buffer)

        label_settings.setFormat(format)

        label_settings.scaleVisibility = True
        label_settings.maximumScale = max_ratio  # Closest zoom level
        label_settings.minimumScale = 1 / min_ratio  # Furthest zoom level

        label_settings.placement = QgsPalLayerSettings.AroundPoint
        label_settings.priority = 5
        label_settings.obstacleScale = 1.0

        layer_settings = QgsVectorLayerSimpleLabeling(label_settings)
        layer.setLabeling(layer_settings)
        layer.setLabelsEnabled(True)


def make_line_symbol(
    culling_mode, edge_color, edge_width, extrusion, facades, offset
) -> Any:
    # ->q3d.QgsPolygon3DSymbol:
    import qgis._3d as q3d

    symbol = q3d.QgsLine3DSymbol()
    symbol.setWidth(edge_width)
    symbol.setOffset(offset)
    symbol.setExtrusionHeight(extrusion)
    symbol.setRenderAsSimpleLines(True)

    return symbol


def make_point_symbol(
    culling_mode, edge_color, edge_width, extrusion, facades, offset
) -> Any:
    # ->q3d.QgsPolygon3DSymbol:
    import qgis._3d as q3d

    symbol = q3d.QgsPoint3DSymbol()
    # symbol.setShape(q3d.QgsSymbol3DShape.Cylinder)
    # symbol.setHeight()
    # symbol.setRadius(extrusion)
    # symbol.setCullingMode(culling_mode.value)
    # symbol.setTransformation(0, 0, offset)
    return symbol


def set_3d_view_settings(
    layers: QgsVectorLayer,
    *,
    offset: float = 0,
    extrusion: float = 4,
    color: TripleNumber = (222, 222, 222),
    facades: Qgis3dFacade = Qgis3dFacade.walls,
    culling_mode: Qgis3dCullingMode = Qgis3dCullingMode.front_face,
    repaint: bool = True,
    edge_width: float = 1.0,
    edge_color: TripleNumber = (255, 255, 255),
) -> None:
    if layers is None:
        return

    polygon_renderer = make_renderer(
        color,
        make_polygon_symbol(
            culling_mode, edge_color, edge_width, extrusion, facades, offset
        ),
    )

    line_renderer = make_renderer(
        color,
        make_line_symbol(
            culling_mode, edge_color, edge_width, extrusion, facades, offset
        ),
    )

    point_renderer = make_renderer(
        color,
        make_point_symbol(
            culling_mode, edge_color, edge_width, extrusion, facades, offset
        ),
    )

    for layers_inner in layers:
        if layers_inner:
            if isinstance(layers_inner, Iterable):
                for layer in layers_inner:
                    if layer:
                        set_renderer(
                            layer, line_renderer, point_renderer, polygon_renderer
                        )

                        if repaint:
                            layer.triggerRepaint()
            else:
                set_renderer(
                    layers_inner, line_renderer, point_renderer, polygon_renderer
                )
                if repaint:
                    layers_inner.triggerRepaint()


def set_renderer(layer, line_renderer, point_renderer, polygon_renderer):
    from qgis.core import QgsWkbTypes

    if layer.geometryType() == QgsWkbTypes.PointGeometry:  # QgsWkbTypes.Point:
        layer.setRenderer3D(point_renderer)
    elif layer.geometryType() == QgsWkbTypes.LineGeometry:  # QgsWkbTypes.Line:
        layer.setRenderer3D(line_renderer)
    elif layer.geometryType() == QgsWkbTypes.PolygonGeometry:  # QgsWkbTypes.Polygon:
        layer.setRenderer3D(polygon_renderer)
    else:
        logger.error(f"geometry type not supported: {layer.geometryType()}, skipping")


def make_renderer(color, symbol):
    import qgis._3d as q3d

    apply_common_symbol_settings(symbol)
    apply_material(color, symbol)
    renderer = q3d.QgsVectorLayer3DRenderer()
    renderer.setSymbol(symbol)

    return renderer


def make_polygon_symbol(
    culling_mode, edge_color, edge_width, extrusion, facades, offset
) -> Any:
    # ->q3d.QgsPolygon3DSymbol:
    import qgis._3d as q3d

    symbol = q3d.QgsPolygon3DSymbol()
    symbol.setCullingMode(culling_mode.value)
    symbol.setOffset(offset)
    symbol.setExtrusionHeight(extrusion)
    symbol.setRenderedFacade(facades.value)

    if edge_width > 0:
        symbol.setEdgesEnabled(True)
    else:
        symbol.setEdgesEnabled(False)

    symbol.setEdgeWidth(edge_width)
    symbol.setEdgeColor(QColor(*edge_color))
    symbol.setAddBackFaces(False)
    # symbol.setInvertNormals(False)

    return symbol


def apply_common_symbol_settings(symbol) -> None:
    if symbol is None:
        logger.error("symbol is None, skipping")
        return

    if hasattr(symbol, "setAltitudeBinding"):
        symbol.setAltitudeBinding(Qgis3dAltitudeBinding.centroid.value)
    if hasattr(symbol, "setAltitudeClamping"):
        symbol.setAltitudeClamping(Qgis3dAltitudeClamping.absolute.value)


def apply_material(color, symbol) -> None:
    import qgis._3d as q3d

    material_settings = q3d.QgsPhongMaterialSettings()

    q_color = QColor(*color)
    material_settings.setAmbient(q_color)
    material_settings.setDiffuse(q_color)
    material_settings.setSpecular(q_color)
    symbol.setMaterialSettings(material_settings)



from typing import Generator, Iterable, List, Optional, Sequence, Tuple, Union

import numpy
import shapely
from shapely.geometry import LineString, MultiPoint, Point
from shapely.geometry.base import BaseGeometry
from warg import Number

__all__ = [
    "unique_line_points",
    "nearest_neighbor_within",
    "azimuth",
    "shift_point",
    "closest_object",
]


def unique_line_points(lines: Sequence[LineString]) -> List[Point]:
    """


    :param lines:
    :return: Return list of unique vertices from list of LineStrings.
    :rtype: List[Point]
    """

    vertices = []

    for line in lines:
        vertices.extend(list(line.coords))

    return [Point(p) for p in set(vertices)]


def nearest_neighbor_within(
    others: Sequence, point: shapely.Point, max_distance: Number
) -> Optional[Point]:
    """Find the nearest point among others up to a maximum distance.


    :param others: a list of Points or a MultiPoint
    :param point: a Point
    :param max_distance: maximum distance to search for the nearest neighbor

    :return: A shapely Point if one is within max_distance, None otherwise
    :rtype: Optional[Point]
    """
    search_region = point.buffer(max_distance)
    interesting_points = search_region.intersection(MultiPoint(others))

    if not interesting_points:
        closest_point = None
    elif isinstance(interesting_points, Point):
        closest_point = interesting_points
    else:
        if isinstance(interesting_points, MultiPoint):
            interesting_points = list(interesting_points.geoms)

        distances = [
            point.distance(ip) for ip in interesting_points if point.distance(ip) > 0
        ]
        closest_point = interesting_points[distances.index(min(distances))]

    return closest_point


def closest_object(
    geometries: Iterable[BaseGeometry], point: Point
) -> Tuple[BaseGeometry, float, int]:
    """
    Find the nearest geometry among a list, measured from fixed point.

    :param geometries:  a iterable of shapely geometry objects
    :param point: a shapely Point
    :return: Tuple (geom, min_dist, min_index) of the geometry with minimum distance
        to point, its distance min_dist and the list index of geom, so that
        geom = geometries[min_index].
    """
    if isinstance(geometries, Generator):
        geometries = list(geometries)

    min_dist, min_index = min(
        (point.distance(geom), k) for (k, geom) in enumerate(geometries)
    )

    return geometries[min_index], min_dist, min_index


def shift_point(
    c1: Union[Point, Tuple[Number, Number]],
    c2: Union[Point, Tuple[Number, Number]],
    offset: float,
) -> Point:
    """

    shift points with offset in orientation of line c1->c2

    :param c1:
    :param c2:
    :param offset:
    :return:
    """

    if isinstance(c1, Point):
        x1, y1 = c1.coords[0]
    else:
        x1, y1 = c1

    if isinstance(c2, Point):
        x2, y2 = c2.coords[0]
    else:
        x2, y2 = c2

    if ((x1 - x2) == 0) and ((y1 - y2) == 0):  # zero length line
        x_new, y_new = x1, y1
    else:
        rel_length = numpy.minimum(
            offset / numpy.sqrt((x1 - x2) ** 2 + (y1 - y2) ** 2), 1
        )
        x_new = x1 + (x2 - x1) * rel_length
        y_new = y1 + (y2 - y1) * rel_length

    return Point(x_new, y_new)


def azimuth(point1: Point, point2: Point) -> float:
    """
    The clockwise angle from North to line of two points

    :param point1:
    :type point1: Point
    :param point2:
    :type point2: Point
    :return: angle
    :rtype: float
    """

    angle = numpy.arctan2(point2.x - point1.x, point2.y - point1.y)
    # Gets the angle between the first and last coordinate of a linestring

    return (
        numpy.degrees(angle) if angle >= 0 else numpy.degrees(angle) + 360
    ) % 180  # Modulo is used on the angle to produce a result between 0 and 180 degrees


if __name__ == "__main__":
    print(azimuth(Point(0, 0), Point(1, 1)))
    print(azimuth(Point(1, 1), Point(0, 0)))

    print(shift_point(Point(1, 1), Point(0, 0), 1))
    print(shift_point(Point(1, 1), Point(0, 0), 2))
    print(shift_point(Point(1, 1), Point(0, 0), 3))

    print(shift_point(Point(0, 0), Point(1, 1), 1))
    print(shift_point(Point(0, 0), Point(1, 1), 0))



import logging
from typing import Any, Callable, List, Mapping, Sequence, Union

import shapely
from shapely import unary_union

from .geometry_types import is_multi
from .morphology import clean_shape, closing

__all__ = ["overlap_groups"]

logger = logging.getLogger(__name__)


def overlap_groups(
    to_be_grouped: Union[
        Sequence[shapely.geometry.base.BaseGeometry],
        Mapping[Any, shapely.geometry.base.BaseGeometry],
    ],
    must_be_unique: bool = False,
    group_test: Callable = shapely.intersects,
) -> List[Mapping[Any, shapely.geometry.base.BaseGeometry]]:
    """

    Given a sequence of geometries `to_be_grouped`, this function will group
    the geometries based on a `group_test` function. `group_test` must return a bool indicating whether to
    group any two geometries

    `must_be_unique` allows us to assert whether all geometries will be possible grouped uniquely, e.g. if in
    `to_be_grouped` a multi shape could end up in 2 groups, we can disallow that.

    :param to_be_grouped:
    :param must_be_unique:
    :param group_test:
    :return:
    """

    if not isinstance(to_be_grouped, Mapping):
        to_be_grouped = dict(zip((i for i in range(len(to_be_grouped))), to_be_grouped))

    if must_be_unique:
        assert not any(is_multi(p) for p in to_be_grouped.values()), to_be_grouped

    unions = clean_shape(
        closing(unary_union(list(unary_union(v) for v in to_be_grouped.values())))
    )

    groups = []
    already_grouped = []

    if not is_multi(unions):
        groups.append(to_be_grouped)
    else:
        for union_part in unions.geoms:
            union_part = clean_shape(union_part)
            incidentee = {}
            for k, v in to_be_grouped.items():
                v = clean_shape(v)
                try:
                    g_test = group_test(v, union_part)
                except shapely.errors.GEOSException as e:
                    logger.error(e)  # Assume overlap was found
                    g_test = True

                if g_test:
                    if must_be_unique:
                        assert k not in already_grouped, f"{k, already_grouped, v}"
                    incidentee[k] = v
                    already_grouped.append(k)
            groups.append(incidentee)

    return groups


if __name__ == "__main__":

    def demo():
        from shapely.geometry import box
        from pprint import pprint

        data = [
            box(1, 1, 3, 3),
            box(2, 2, 3, 3),
            box(4, 4, 6, 6),
            box(4, 4, 5, 5),
            box(5, 5, 6, 6),
            box(7, 7, 8, 8),
            box(1, 1, 2, 2),
            box(4, 4, 6, 6),
        ]

        pprint(overlap_groups(data))

        data = [
            box(1, 1, 3, 3),
            unary_union([box(2, 2, 3, 3), box(4, 4, 5, 5)]),
            box(4, 4, 6, 6),
            box(4, 4, 5, 5),
        ]

        pprint(overlap_groups(data))

        # pprint(overlap_groups(data, must_be_unique=True)) # FAILS!

    demo()



from enum import Enum

from qgis._3d import Qgs3DTypes

# noinspection PyUnresolvedReferences
from qgis.core import (
    QgsMultiBandColorRenderer,
    QgsPalettedRasterRenderer,
    QgsSingleBandColorDataRenderer,
    QgsSingleBandGrayRenderer,
    QgsSingleBandPseudoColorRenderer,
)

from jord.geojson_utilities import GeoJsonGeometryTypesEnum

__all__ = [
    "QgisRendererEnum",
    "QgisLayerTypeEnum",
    "Qgis3dCullingMode",
    "Qgis3dFacade",
    "Qgis3dAltitudeBinding",
    "Qgis3dAltitudeClamping",
]


class QgisRendererEnum(Enum):
    multi_band = QgsMultiBandColorRenderer
    paletted_raster = QgsPalettedRasterRenderer
    single_band_color = QgsSingleBandColorDataRenderer
    single_band_gray = QgsSingleBandGrayRenderer
    single_band_pseudo = QgsSingleBandPseudoColorRenderer


class QgisLayerTypeEnum(Enum):
    """
    This enum is useful for exhaustively iterating possible GeoJson types.

    NOTE: Z coordinate support is not included in this enum.
    """

    point = GeoJsonGeometryTypesEnum.point.value.__name__
    multi_point = GeoJsonGeometryTypesEnum.multi_point.value.__name__
    line_string = GeoJsonGeometryTypesEnum.line_string.value.__name__
    multi_line_string = GeoJsonGeometryTypesEnum.multi_line_string.value.__name__
    polygon = GeoJsonGeometryTypesEnum.polygon.value.__name__
    multi_polygon = GeoJsonGeometryTypesEnum.multi_polygon.value.__name__
    curve_polygon = "CurvePolygon"
    multi_surface = "MultiSurface"
    compound_curve = "CompoundCurve"
    multi_curve = "MultiCurve"
    no_geometry = "No Geometry"


class Qgis3dCullingMode(Enum):
    no_culling = 0  # Qgs3DTypes.CullingMode.NoCulling # 0
    front_face = 1
    back_face = 2
    front_and_back_face = 3


class Qgis3dAltitudeBinding(Enum):
    vertex = (
        Qgs3DTypes.AltitudeBinding.Vertex
    )  # 0  # Vertex: Clamp every vertex of feature

    centroid = (
        Qgs3DTypes.AltitudeBinding.Centroid
    )  # Centroid: Clamp just centroid of feature


class Qgis3dAltitudeClamping(Enum):
    absolute = (
        Qgs3DTypes.AltitudeClamping.Absolute
    )  # 0  # Absolute: Elevation is taken directly from feature
    # and is independent of terrain height (
    # final elevation = feature elevation)

    relative = (
        Qgs3DTypes.AltitudeClamping.Relative
    )  # 1  # Relative: Elevation is relative to terrain height
    # (final elevation = terrain elevation +
    # feature elevation)

    terrain = (
        Qgs3DTypes.AltitudeClamping.Terrain
    )  # 2  # Terrain: Elevation is clamped to terrain (final
    # elevation = terrain elevation)


class Qgis3dFacade(Enum):
    no_facade = 0
    walls = 1
    roofs = 2
    walls_and_roofs = 3



import statistics
from typing import Generator, Iterable, List, Sequence, Tuple, Union

import shapely
from shapely import MultiPolygon, Polygon
from shapely.geometry.base import BaseGeometry
from warg import Number, pairs

from .base import sanitise
from .geometry_types import is_multi
from .lines import segments

__all__ = [
    "explode_polygons",
    "polygon_has_interior_rings",
    "iter_polygons",
    "discard_holes",
    "get_coords_from_polygonal_shape",
    "get_polygonal_shape_from_coords",
    "has_holes",
    "is_polygonal",
]

DEFAULT_DISTANCE = 1e-7


def polygon_has_interior_rings(polygon: shapely.Polygon) -> bool:
    """

    :param polygon:
    :return:
    """
    return len(polygon.interiors) > 0


def mean_std_dev_line_length(geom: BaseGeometry) -> Tuple[float, float]:
    """

    :return:
    :param geom:
    :return:
    """
    line_lengths = []
    if isinstance(geom, shapely.LineString):
        for segment in segments(geom):
            line_lengths.append(segment.length)
    elif isinstance(geom, shapely.MultiLineString):
        for li in geom.geoms:
            for segment in segments(li):
                line_lengths.append(segment.length)
    elif isinstance(geom, (shapely.Polygon, shapely.MultiPolygon)):
        exterior_rings, interior_rings = extract_poly_rings(geom)
        for ex_ring in exterior_rings:
            for segment in segments(ex_ring):
                line_lengths.append(segment.length)
        for in_ring in interior_rings:
            for segment in segments(in_ring):
                line_lengths.append(segment.length)
    else:
        raise ValueError(f"Unhandled geometry type: {repr(geom.type)}")

    return (
        statistics.mean(line_lengths),
        statistics.stdev(line_lengths) if len(line_lengths) > 1 else 0,
    )


def mean_std_dev_area(geom: BaseGeometry) -> Tuple[float, float]:
    """

    :param geom:
    :return:
    """
    poly_areas = []
    if isinstance(geom, shapely.Polygon):
        poly_areas.append(geom.area)
    elif isinstance(geom, shapely.MultiPolygon):
        for po in geom.geoms:
            poly_areas.append(po.area)
    else:
        raise ValueError(f"Unhandled geometry type: {repr(geom.type)}")

    return (
        statistics.mean(poly_areas),
        statistics.stdev(poly_areas) if len(poly_areas) > 1 else 0,
    )


def prune_area(geom: BaseGeometry, eps: float = DEFAULT_DISTANCE) -> BaseGeometry:
    """

    :param geom:
    :param eps:
    :return:
    """
    raise NotImplementedError
    poly_areas = []
    if isinstance(geom, Polygon):
        poly_areas.append(geom.area)

    elif isinstance(geom, MultiPolygon):
        for po in geom.geoms:
            poly_areas.append(po.area)

    return poly_areas


def prune_rings(geom: BaseGeometry, eps: float = DEFAULT_DISTANCE) -> BaseGeometry:
    """

    :param geom:
    :param eps:
    :return:
    """
    raise NotImplementedError
    poly_areas = []
    if isinstance(geom, Polygon):
        poly_areas.append(geom.area)
    elif isinstance(geom, MultiPolygon):
        for po in geom.geoms:
            poly_areas.append(po.area)

    return poly_areas


def prune_holes(
    geom: Union[shapely.MultiPolygon, shapely.Polygon], epsilon: Number = 1000
) -> Union[shapely.MultiPolygon, shapely.Polygon]:
    """

    :param epsilon:
    :param geom:
    :return:"""

    if isinstance(geom, shapely.MultiPolygon):
        parts = []

        for polygon in geom.geoms:
            interiors = []

            for interior in polygon.interiors:
                p = shapely.Polygon(interior)

                if p.area > epsilon:
                    interiors.append(interior)

            temp_pol = shapely.Polygon(polygon.exterior.coords, holes=interiors)
            parts.append(temp_pol)

        return shapely.MultiPolygon(parts)

    interiors = []

    for interior in geom.interiors:
        p = shapely.Polygon(interior)
        if p.area > epsilon:
            interiors.append(interior)

    return shapely.Polygon(geom.exterior.coords, holes=interiors)


def get_coords_from_polygonal_shape(
    shape: Union[shapely.Polygon, shapely.MultiPolygon],
) -> List[List[List[tuple[float, float]]]]:
    coords = []

    if isinstance(shape, shapely.Polygon):
        coords.append(shape.exterior.coords[:])
        for linearring in shape.interiors:
            coords.append(linearring.coords[:])
    elif isinstance(shape, shapely.MultiPolygon):
        for polygon in shape.geoms:
            coords.append(get_coords_from_polygonal_shape(polygon))

    return coords


def get_polygonal_shape_from_coords(
    coords: Union[
        Iterable[Iterable[Iterable[tuple[float, float]]]],
        Iterable[Iterable[tuple[float, float]]],
    ],
) -> Union[shapely.Polygon, shapely.MultiPolygon, None]:
    outer = next(iter(coords), None)

    assert isinstance(outer, Iterable)
    a = next(iter(outer), None)
    if isinstance(a, Iterable):  # Polygon
        b = next(iter(a), None)
        if isinstance(b, Iterable):  # Holes
            polygons = []
            for poly in coords:  # MultiPolygon and # MultiPolygon Holes
                polygons.append(get_polygonal_shape_from_coords(poly))
            return shapely.MultiPolygon(polygons)
        else:
            exterior, *interior = coords
            if interior:
                return shapely.Polygon(exterior, holes=interior)
            return shapely.Polygon(exterior)

    if len(coords[0]) == 0:
        return None

    return shapely.Polygon(coords)


def extract_poly_coords(
    geom: Union[shapely.Polygon, shapely.MultiPolygon],
) -> Tuple[List, List]:
    """
    TODO: Duplicate of get_coords_from_polygonal_shape

    :param geom:
    :return:
    """
    if geom.type == "Polygon":
        exterior_coords = geom.exterior.coords[:]
        interior_coords = []
        for interior in geom.interiors:
            interior_coords += interior.coords[:]
    elif geom.type == "MultiPolygon":
        exterior_coords = []
        interior_coords = []
        for part in geom.geoms:
            epc = extract_poly_coords(part)  # Recursive call
            exterior_coords += epc[0]
            interior_coords += epc[1]
    else:
        raise ValueError(f"Unhandled geometry type: {repr(geom.type)}")
    return exterior_coords, interior_coords


def extract_poly_rings(geom: BaseGeometry) -> Tuple[List, List]:
    """

    :param geom:
    :return:
    """
    interior_rings = []
    exterior_rings = []
    if isinstance(geom, shapely.Polygon):
        exterior_rings.append(geom.exterior)
        interior_rings.extend(geom.interiors)
    elif isinstance(geom, shapely.MultiPolygon):
        for part in geom.geoms:
            exterior_rings.append(part.exterior)
            interior_rings.extend(part.interiors)
    else:
        raise ValueError(f"Unhandled geometry type: {repr(geom.type)}")

    return exterior_rings, interior_rings


def discard_holes(
    shape: Union[shapely.Polygon, shapely.MultiPolygon],
) -> Union[shapely.Polygon, shapely.MultiPolygon]:
    if isinstance(shape, shapely.Polygon):
        return shapely.Polygon(shape.exterior.coords)

    elif isinstance(shape, shapely.MultiPolygon):
        shape_parts = []
        for shape_part in shape.geoms:
            shape_parts.append(shapely.Polygon(shape_part.exterior.coords))
        return shapely.MultiPolygon(shape_parts)

    elif isinstance(shape, shapely.GeometryCollection):
        shape_parts = []
        for shape_part in shape.geoms:
            if isinstance(shape_part, (shapely.Polygon, shapely.MultiPolygon)):
                shape_parts.append(discard_holes(shape_part))
            else:
                shape_parts.append(shape_part)
        return shapely.GeometryCollection(shape_parts)

    raise NotImplementedError(
        f"Discarding hole for "
        f"{shape.type if isinstance(shape, shapely.geometry.base.BaseGeometry) else type(shape)} is not "
        f"implemented"
    )


def has_holes(shape: Union[shapely.Polygon, shapely.MultiPolygon]) -> bool:
    if is_multi(shape):
        return any(has_holes(s) for s in shape.geoms)

    if isinstance(shape, shapely.Polygon):
        return len(shape.interiors) > 0

    # raise #not polygonal
    return False


def is_polygonal(cleaned):
    if isinstance(
        cleaned,
        (
            shapely.Point,
            shapely.MultiPoint,
            shapely.LineString,
            shapely.MultiLineString,
        ),
    ):
        return False
    elif isinstance(cleaned, shapely.GeometryCollection):
        return any(is_polygonal(p) for p in cleaned.geoms)
    return True


def iter_polygons(
    _input_geometry: BaseGeometry,
) -> Union[Generator[shapely.Polygon, None, None], Tuple[BaseGeometry]]:
    """

    :param _input_geometry:
    :return:
    """
    if isinstance(_input_geometry, shapely.MultiPolygon):
        return (polygon for polygon in _input_geometry.geoms)
    elif isinstance(_input_geometry, shapely.GeometryCollection):
        return (poly for poly in _input_geometry.geoms if is_polygonal(poly))
    elif isinstance(_input_geometry, Iterable):
        return (
            y
            for x in (
                poly.geoms if is_multi(poly) else iter_polygons(poly)
                for poly in _input_geometry
                if is_polygonal(poly)
            )
            for y in x
        )

    assert isinstance(_input_geometry, shapely.Polygon)

    return (_input_geometry,)


def explode_polygons(
    polygons: Union[shapely.Polygon, shapely.MultiPolygon, Sequence[shapely.Polygon]],
    return_index: bool = False,
) -> Union[
    Sequence[shapely.LineString], Tuple[Sequence[shapely.LineString], Sequence[int]]
]:
    """

    :param polygons:
    :param return_index:
    :return: main line features that make up the polygons
    """
    lines_out = []
    index = []

    if isinstance(polygons, shapely.Polygon):
        polygons = [polygons]

    if isinstance(polygons, shapely.MultiPolygon):
        polygons = polygons.geoms

    for i, l in enumerate(polygons):
        for s in [shapely.LineString(s) for s in pairs(l.exterior.coords)]:
            lines_out.append(s)
            index.append(i)

        for p in l.interiors:
            lines_out.append(shapely.LineString(p.coords))
            index.append(i)

    if return_index:
        return lines_out, index

    return lines_out


if __name__ == "__main__":

    def aishdjauisd():
        # Import constructors for creating geometry collections
        from shapely.geometry import MultiPoint, MultiLineString

        # Import necessary geometric objects from shapely module
        from shapely.geometry import Point, LineString, Polygon

        # Create Point geometric object(s) with coordinates
        point1 = Point(2.2, 4.2)
        point2 = Point(7.2, -25.1)
        point3 = Point(9.26, -2.456)
        # point3D = Point(9.26, -2.456, 0.57)

        # Create a MultiPoint object of our points 1,2 and 3
        multi_point = MultiPoint([point1, point2, point3])

        # It is also possible to pass coordinate tuples inside
        multi_point2 = MultiPoint([(2.2, 4.2), (7.2, -25.1), (9.26, -2.456)])

        # We can also create a MultiLineString with two lines
        line1 = LineString([point1, point2])
        line2 = LineString([point2, point3])
        multi_line = MultiLineString([line1, line2])
        polygon = Polygon([point2, point1, point3])

        from shapely.geometry import GeometryCollection
        from matplotlib import pyplot
        import geopandas

        geoms = GeometryCollection([multi_point, multi_point2, multi_line, polygon])
        print(mean_std_dev_line_length(geoms))
        geoms = sanitise(geoms)
        print(mean_std_dev_line_length(geoms))

        p = geopandas.GeoSeries(geoms)
        p.plot()
        pyplot.show()

    def ahsudh():
        # Import constructors for creating geometry collections
        from shapely.geometry import MultiPoint, MultiLineString

        # Import necessary geometric objects from shapely module
        from shapely.geometry import Point, LineString, Polygon

        # Create Point geometric object(s) with coordinates
        point1 = Point(2.2, 4.2)
        point2 = Point(7.2, -25.1)
        point3 = Point(9.26, -2.456)
        # point3D = Point(9.26, -2.456, 0.57)

        # Create a MultiPoint object of our points 1,2 and 3
        multi_point = MultiPoint([point1, point2, point3])

        # It is also possible to pass coordinate tuples inside
        multi_point2 = MultiPoint([(2.2, 4.2), (7.2, -25.1), (9.26, -2.456)])

        # We can also create a MultiLineString with two lines
        line1 = LineString([point1, point2])
        line2 = LineString([point2, point3])
        multi_line = MultiLineString([line1, line2])
        polygon = Polygon([point2, point1, point3])
        polygon2 = Polygon([point3, point2, point1])
        multi_polygon = shapely.MultiPolygon([polygon, polygon2])

        print(mean_std_dev_line_length(line1))
        print(mean_std_dev_line_length(line2))
        print(mean_std_dev_line_length(multi_line))
        print(mean_std_dev_line_length(polygon))
        print(mean_std_dev_line_length(polygon2))
        print(mean_std_dev_line_length(multi_polygon))

        print(mean_std_dev_area(polygon))
        print(mean_std_dev_area(polygon2))
        print(mean_std_dev_area(multi_polygon))

        from shapely.geometry import GeometryCollection
        from matplotlib import pyplot
        import geopandas

        geoms = GeometryCollection(
            [multi_point, multi_point2, multi_line, polygon, polygon2, multi_polygon]
        )

        p = geopandas.GeoSeries(geoms)
        p.plot()
        pyplot.show()

    # ahsudh()
    # aishdjauisd()



import csv
from pathlib import Path
from typing import Generator, Tuple

__all__ = ["read_gcp_file"]


def read_gcp_file(
    gcp_points_file_path: Path, filter_comments: bool = True
) -> Tuple[Generator, Generator]:
    with open(gcp_points_file_path, encoding="utf8", errors="ignore") as fp:
        if filter_comments:
            fp = filter(lambda row: row[0] != "#", fp)

        gcps = list(csv.DictReader(fp, delimiter=","))

    source_xy, dest_xy = zip(
        *(
            (
                (float(d["sourceX"]), float(d["sourceY"])),
                (float(d["mapX"]), float(d["mapY"])),
            )
            for d in gcps
            if int(d["enable"]) > 0
        )
    )
    return source_xy, dest_xy


if __name__ == "__main__":
    source_xy, dest_xy = read_gcp_file(r"C:\Users\chen\Documents\asdad.gpkg.points")

    print(list(source_xy), list(dest_xy))



__author__ = "heider"
__doc__ = r"""

           Created on 5/5/22
           """

from pathlib import Path

with open(Path(__file__).parent / "README.md") as this_init_file:
    __doc__ += this_init_file.read()

try:
    from .geometry_types import *


except ImportError as ix:
    this_package_name = Path(__file__).parent.name
    print(f"Make sure geopandas module is available for {this_package_name}")
    raise ix



__author__ = "heider"
__doc__ = r"""

           Created on 1/23/23
           """

__all__ = [
    "split_line_string",
    "combine_line_string",
    "extend_segment",
    "fix_starting_point",
    "adjust_line_end",
    "ensure_list_of_geometries",
]

from collections import deque
from typing import List, Sequence, Union

from shapely.geometry import LineString, MultiLineString, Point, Polygon
from shapely.geometry.base import BaseGeometry, BaseMultipartGeometry


def split_line_string(line_string: LineString) -> Sequence[LineString]:
    """
    Break a LineString

    :param line_string:
    :return:
    """
    for start, end in zip(line_string.coords[:-1], line_string.coords[1:]):
        yield LineString((start, end))


def combine_line_string(segments: Sequence[LineString]) -> LineString:
    """

    :param segments:
    :return:
    :rtype: LineString
    """
    coords = [segments[0].coords[0]]

    for segment in segments:
        coords.extend(segment.coords[1:])

    return LineString(coords)


def extend_segment(line_string: LineString) -> LineString:
    """
    Move a line segment's start and end away from each other to ensure intersections

    :param line_string:
    :return:
    """
    p0 = line_string.coords[0]
    p1 = line_string.coords[1]

    difference = (p0[0] - p1[0], p0[1] - p1[1])
    length = (difference[0] ** 2 + difference[1] ** 2) ** 0.5
    direction = (difference[0] / length, difference[1] / length)
    offset = (direction[0] * 1e-12, direction[1] * 1e-12)

    new_p0 = (p0[0] + offset[0], p0[1] + offset[1])
    new_p1 = (p1[0] - offset[0], p1[1] - offset[1])

    return LineString((new_p0, new_p1))


def fix_starting_point(polygon_pieces: Sequence[Polygon]) -> Sequence[Polygon]:
    """
    Reconnect the starting point of a polygon's pieces.
    When splitting a polygon with two lines, we want to get two pieces.
    However, that's not quite how Shapely works.  The outline of the
    polygon is a LinearRing that starts and ends at the same place, but
    Shapely still knows where that starting point is and splits there
    too.
    We don't want that third piece, so we'll reconnect the segments that
    touch the starting point.


    :param polygon_pieces:
    :return:
    """

    if len(polygon_pieces) == 3:
        # Fortunately, Shapely keeps the starting point of the LinearRing
        # as the starting point of the first segment.  That means it's also
        # the ending point of the last segment.  Reconnecting is super simple:
        return [
            polygon_pieces[1],
            LineString(polygon_pieces[2].coords[:] + polygon_pieces[0].coords[1:]),
        ]
    else:
        return polygon_pieces  # We probably cut exactly at the starting point.


def adjust_line_end(line: LineString, end: BaseGeometry) -> LineString:
    """
    Reverse line if necessary to ensure that it ends near end.

    :param line:
    :param end:
    :return:
    """

    line_start = Point(*line.coords[0])
    line_end = Point(*line.coords[-1])

    if line_end.distance(end) < line_start.distance(end):
        return line
    else:
        return LineString(line.coords[::-1])


def ensure_list_of_geometries(
    thing: Union[BaseGeometry, BaseMultipartGeometry],
) -> List[BaseGeometry]:
    """

    :param thing:
    :return:
    """
    if False:
        try:  # Not MultiGeometry base class for shapely
            return list(thing.geoms)
        except AttributeError:
            return [thing]
    if isinstance(thing, BaseMultipartGeometry):
        return list(thing.geoms)
    return [thing]


def clamp_linestring_to_polygon(
    line_string: LineString, polygon: Polygon
) -> LineString:
    """

    :param line_string:
    :param polygon:
    :return:
    """
    segments = deque(split_line_string(line_string))
    result = []
    exiting_segment = None
    was_inside = False

    # contains() checks can fail without this.
    buffered_polygon = polygon.buffer(1e-9)

    while segments:
        current_segment = segments.popleft()
        pieces = ensure_list_of_geometries(current_segment.difference(polygon.exterior))

        if pieces[0].coords[0] != current_segment.coords[0]:
            # The initial part of this line segment coincided with part of the
            # polygon border and was removed.

            if was_inside:
                # If we were already inside, we include this border segment.
                result.append(
                    LineString((current_segment.coords[0], pieces[0].coords[0]))
                )

            # Push the rest back on to be processed later.
            # Note that extendleft() reverses its arguments, so we have to compensate.
            segments.extendleft(reversed(pieces))

        elif len(pieces) > 1 and pieces[0].coords[-1] != pieces[1].coords[0]:
            # There's an initial segment, then a portion that coincided with
            # part of the polygon border.  Break this segment apart and re-process.

            segments.appendleft(
                LineString((pieces[0].coords[1], current_segment.coords[-1]))
            )
            segments.appendleft(pieces[0])

        elif len(pieces) == 1:
            # This segment is either all inside or all outside the polygon.

            is_inside = buffered_polygon.contains(current_segment)
            if is_inside and not was_inside:
                # We've crossed from outside to inside exactly at the starting
                # point of this line segment.

                if exiting_segment:
                    # Earlier we crossed out from the inside, now we're
                    # crossing back in.  Find the shortest path along
                    # the border that gets from the exit point to the
                    # entry point and add it to the result.

                    entering_segment = extend_segment(current_segment)

                    difference = polygon.boundary.difference(
                        MultiLineString((exiting_segment, entering_segment))
                    )

                    polygon_pieces = ensure_list_of_geometries(difference)
                    polygon_pieces = fix_starting_point(polygon_pieces)

                    if len(polygon_pieces) == 1:
                        # We re-entered exactly where we left, so we
                        # don't include any of the border.
                        pass
                    else:
                        shorter = min(polygon_pieces, key=lambda piece: piece.length)

                        # We don't know which direction the polygon border
                        # piece should be.  adjust_line_end() will figure
                        # that out.
                        result.append(
                            adjust_line_end(shorter, Point(*current_segment.coords[0]))
                        )

                    exiting_segment = None

                result.append(current_segment)
            elif was_inside and not is_inside:
                # Like the previous case, but we've crossed to outside.  Store
                # an extended version of this segment as the exit point.
                exiting_segment = extend_segment(current_segment)
            elif is_inside:
                result.append(current_segment)

            was_inside = is_inside

        elif len(pieces) > 1:
            # This segment crosses the border, or touches the border at a single point,
            # or maybe crosses multiple times, etc.  Split the first portion off and
            # push it and the rest to be reprocessed in the following iterations.

            segments.appendleft(
                LineString((pieces[0].coords[1], current_segment.coords[-1]))
            )
            segments.appendleft(pieces[0])

    return combine_line_string(result)


if __name__ == "__main__":

    def _main():
        from draugr.numpy_utilities import HilbertCurve
        from geopandas import GeoSeries
        from matplotlib import pyplot

        hilbert_curve = HilbertCurve(5, 2)
        hc = LineString(hilbert_curve.points_from_distances(range(1023)))
        circle = Point(15, 15).buffer(15)
        clamped = clamp_linestring_to_polygon(hc, circle)

        gs = GeoSeries([clamped])
        gs.plot()
        pyplot.show()

    _main()



__author__ = "heider"
__doc__ = r"""

           Created on 5/5/22
           """

__all__ = [
    "import_qgis",
    # "QGIS"
]

from types import ModuleType


def import_qgis() -> ModuleType:
    """

    :return:
    :rtype: ModuleType
    """
    try:
        import qgis

    except (ImportError, ModuleNotFoundError) as e:
        raise ImportError(f"gdal is not installed {type(e), e}")

    return qgis


# QGIS = import_qgis()



from enum import Enum

from geojson import (
    Point,
    LineString,
    Polygon,
    MultiPoint,
    MultiLineString,
    MultiPolygon,
    GeometryCollection,
)

__all__ = ["GeoJsonGeometryTypesEnum"]


class GeoJsonGeometryTypesEnum(Enum):
    """
    This enum is useful for exhaustively iterating possible geojson types.
    """

    point = Point  #

    line_string = LineString  #

    polygon = Polygon  #

    multi_point = MultiPoint  #

    multi_line_string = MultiLineString  #

    multi_polygon = MultiPolygon  #

    geometry_collection = GeometryCollection  #


if __name__ == "__main__":
    print(GeoJsonGeometryTypesEnum.geometry_collection.value.__name__)



import shapely


def add_polygon_z(polygon_2d, z_s):
    poly_exterior_s = polygon_2d.exterior.coords[:-1]
    assert len(poly_exterior_s) == len(z_s), (len(poly_exterior_s), len(z_s))

    polygon_3d = shapely.Polygon(
        [(x, y, z) for (x, y), z in zip(poly_exterior_s, z_s, strict=True)]
    )

    return polygon_3d


def add_linestring_z(linestring_2d, z_s):
    linestring_s = linestring_2d.coords
    assert len(linestring_s) == len(z_s), (len(linestring_s), len(z_s))

    linestring_3d = shapely.LineString(
        [(x, y, z) for (x, y), z in zip(linestring_s, z_s, strict=True)]
    )

    return linestring_3d


def add_point_z(point_2d, z):
    return shapely.Point([point_2d.x, point_2d.y, z])


def add_multipoint_z(multipoint_2d, z_s):
    return shapely.MultiPoint(
        [(x, y, z) for (x, y), z in zip(multipoint_2d, z_s, strict=True)]
    )


def add_multilinestring_z(multilinestring_2d, z_s):
    return shapely.MultiLineString(
        [
            [(x, y, z) for (x, y), z in zip(linestring.coords, z_s, strict=True)]
            for linestring in multilinestring_2d
        ]
    )


def add_multipolygon_z(multipolygon_2d, z_s):
    return shapely.MultiPolygon(
        [
            [
                [
                    (x, y, z)
                    for (x, y), z in zip(polygon.exterior.coords, z_s, strict=True)
                ]
                for polygon in multipolygon_2d
            ]
        ]
    )


def add_geometrycollection_z(geometrycollection_2d, z_s):
    return shapely.GeometryCollection(
        [
            add_geometry_z(geometry, z)
            for geometry, z in zip(geometrycollection_2d, z_s, strict=True)
        ]
    )


def add_geometry_z(geometry, z):
    if isinstance(geometry, shapely.geometry.Point):
        return add_point_z(geometry, z)
    elif isinstance(geometry, shapely.geometry.LineString):
        return add_linestring_z(geometry, z)
    elif isinstance(geometry, shapely.geometry.Polygon):
        return add_polygon_z(geometry, z)
    elif isinstance(geometry, shapely.geometry.MultiPoint):
        return add_multipoint_z(geometry, z)
    elif isinstance(geometry, shapely.geometry.MultiLineString):
        return add_multilinestring_z(geometry, z)
    elif isinstance(geometry, shapely.geometry.MultiPolygon):
        return add_multipolygon_z(geometry, z)
    elif isinstance(geometry, shapely.geometry.GeometryCollection):
        return add_geometrycollection_z(geometry, z)
    else:
        raise NotImplementedError(geometry)


def strip_z(geometry):  # TODO: IMPLEMENT
    if isinstance(geometry, shapely.geometry.Point):
        return shapely.Point([geometry.x, geometry.y])
    elif isinstance(geometry, shapely.geometry.LineString):
        return shapely.LineString(geometry.coords.xy)
    elif isinstance(geometry, shapely.geometry.Polygon):
        return shapely.Polygon(geometry.exterior.coords.xy)
    elif isinstance(geometry, shapely.geometry.MultiPoint):
        return shapely.MultiPoint([point.coords.xy for point in geometry])
    elif isinstance(geometry, shapely.geometry.MultiLineString):
        return shapely.MultiLineString(
            [linestring.coords.xy for linestring in geometry]
        )
    elif isinstance(geometry, shapely.geometry.MultiPolygon):
        return shapely.MultiPolygon(
            [polygon.exterior.coords.xy for polygon in geometry]
        )
    elif isinstance(geometry, shapely.geometry.GeometryCollection):
        return shapely.GeometryCollection([strip_z(geometry) for geometry in geometry])
    else:
        raise NotImplementedError(geometry)


if __name__ == "__main__":

    def iuasjhd():
        print(
            add_polygon_z(
                shapely.geometry.Polygon([(0, 0), (1, 0), (1, 1), (0, 1)]), [0, 1, 2, 3]
            )
        )

        print(add_linestring_z(shapely.geometry.LineString([(0, 0), (1, 1)]), [0, 1]))

        print(add_point_z(shapely.geometry.Point([0, 0]), 0))

        # print(strip_z(add_polygon_z(shapely.geometry.Polygon([(0, 0), (1, 0), (1, 1), (0, 1)]), [0, 1, 2, 3])))

    iuasjhd()



__author__ = "Christian Heider Lindbjerg"

import sys


def test_sanity():
    assert True
    assert False is not True
    answer_to_everything = str(42)
    assert str(42) == answer_to_everything


def test_print(capsys):
    """Correct my_name argument prints"""
    text = "hello"
    err = "world"
    print(text)
    sys.stderr.write("world")
    captured = capsys.readouterr()
    assert text in captured.out
    assert err in captured.err



import datetime
from typing import Any, List, Optional

ADD_STRING_LEN = True
NUM_MB16_CHARS = 16777216

__all__ = ["solve_type", "solve_type_configuration"]


def solve_type(d: Any) -> str:
    """
    Does not support size/length yet...

    QGUS Availible Field types:
    ________________           - Provider type ( MemoryLayer) - implemented

    Whole Number (integer) - integer - X
    Decimal Number (real) - double - X
    Text (string) - string - X
    Date - date - X
    Time - time - X
    Date & Time - datetime - X
    Whole Number ( ... llint - 16bit) - int2 - O
    Whole Number (integer - 32bit) - int4 - O
    Whole Number (integer - 64bit) - int8 - O
    Decimal Number (numeric) - numeric - O
    Decimal Number (decimal) - decimal - O
    Decimal Number (real) - real - O
    Decimal Number (double) - double precision - O
    Text, unlimited length (text) - text - X
    Boolean - boolean  - X
    Binary Object (BLOB) - binary - X
    String List - stringlist - X
    Integer List - integerlist - O
    Decimal (double) List - doublelist - O
    Integer (64 bit) List - integer64list - O
    Map - map - O
    Geometry - geometry - O

    :param d:
    :return:
    """
    if not isinstance(d, bool):
        if isinstance(d, int):
            return "integer"

        elif isinstance(d, float):
            return "double"

        elif isinstance(d, bytes):
            return "binary"

        elif isinstance(d, (list, tuple, set)):  # ASSUME IS STRINGS
            return "stringlist"

        elif isinstance(d, datetime.datetime):
            return "datetime"

        elif isinstance(d, datetime.date):
            return "date"

        elif isinstance(d, datetime.time):
            return "time"

        elif isinstance(d, str):
            if False:
                if (
                    ADD_STRING_LEN
                ):  # WARNING! Shapefiles have a limitation of maximum 254 characters per field
                    return f"string({min(max(len(d) * 16, 255), NUM_MB16_CHARS)})"  # 16x buffer for large strings
            else:
                return "text"

    if isinstance(d, bool):
        if False:
            if ADD_STRING_LEN:
                return "string(255)"  # True, False (5)
        else:
            return "boolean"

    if False:
        return "text"

    return "string"


def solve_type_configuration(
    d: Any,
    k: Optional[str],
    columns: Optional[List],
    allocation_multiplier: Optional[int] = 2,
) -> Optional[str]:
    """

    :param d:
    :param k:
    :param columns:
    :param allocation_multiplier:
    :return:
    """
    if isinstance(d, str):
        a = len(d)

        if k and columns:
            max_len = a

            if isinstance(columns, List):
                for cols in columns:
                    c = cols[k]
                    if isinstance(c, str):
                        max_len = max(max_len, len(c))

            a = max_len

        if allocation_multiplier:
            a *= allocation_multiplier

        a = max(a, 255)
        return str(a)

    return None



import logging

from shapely import Polygon

from .rings import ensure_ccw_ring, ensure_cw_ring

logger = logging.getLogger(__name__)

__all__ = [
    "ensure_cw_poly",
    "ensure_ccw_poly",
]


def ensure_ccw_poly(polygon: Polygon) -> Polygon:
    """
    This function checks if the polygon is counter-clockwise if not it is reversed


    :param polygon: The polygon to check
    :return: Returns the polygon turned clockwise
    """

    return Polygon(
        shell=ensure_ccw_ring(polygon.exterior),
        holes=[ensure_ccw_ring(hole) for hole in polygon.interiors],
    )


def ensure_cw_poly(polygon: Polygon) -> Polygon:
    """
    This function checks if the polygon is clockwise if not it is reversed


    :param polygon: The polygon to check
    :return: Returns the polygon turned clockwise
    """

    return Polygon(
        shell=ensure_cw_ring(polygon.exterior),
        holes=[ensure_cw_ring(hole) for hole in polygon.interiors],
    )



import logging
from typing import Optional

import shapely
from shapely.geometry.base import BaseGeometry

__all__ = [
    "closing",
    "opening",
    "erode",
    "erosion",
    "dilate",
    "dilation",
    "close",
    "clean_shape",
    "zero_buffer",
    "pro_opening",
    "pro_closing",
    "BecameEmptyException",
    "collapse_duplicate_vertices",
]

from shapely.validation import make_valid

from warg import passes_kws_to

from .uniformity import ensure_cw_poly

FALLBACK_LINESTRING_CAPSTYLE = shapely.BufferCapStyle.square  # CAN BE OVERRIDDEN
FALLBACK_POINT_CAPSTYLE = shapely.BufferCapStyle.round  # CAN BE OVERRIDDEN
FALLBACK_POINT_JOINSTYLE = shapely.BufferCapStyle.round  # CAN BE OVERRIDDEN
DEFAULT_DISTANCE = 1e-7

logger = logging.getLogger(__name__)


@passes_kws_to(shapely.geometry.base.BaseGeometry.buffer)
def morphology_buffer(
    geom: BaseGeometry,
    distance: float = DEFAULT_DISTANCE,
    cap_style: shapely.BufferCapStyle = shapely.BufferCapStyle.flat,
    join_style: shapely.BufferJoinStyle = shapely.BufferJoinStyle.mitre,
    **kwargs,
):
    if geom.is_empty:
        return geom

    if distance == 0:
        if isinstance(geom, shapely.GeometryCollection):
            return shapely.GeometryCollection(
                [
                    morphology_buffer(
                        g,
                        distance=distance,
                        cap_style=cap_style,
                        join_style=join_style,
                        **kwargs,
                    )
                    for g in geom.geoms
                ]
            )

        if not isinstance(
            geom, (shapely.Polygon, shapely.MultiPolygon)
        ):  # So if line(s) or point(s)
            return geom

    if isinstance(
        geom, (shapely.Polygon, shapely.MultiPolygon, shapely.GeometryCollection)
    ):
        if geom.area == 0:
            geom_ = geom.boundary
            if geom_:
                geom = geom_

    if isinstance(geom, (shapely.LineString, shapely.MultiLineString)):
        if geom.length == 0:
            geom_ = geom.representative_point()
            if geom_:
                geom = geom_

    if (
        cap_style == shapely.BufferCapStyle.flat
    ):  # test for parameter NONSENSE, probably not what was intended
        if isinstance(geom, (shapely.Point, shapely.MultiPoint)):
            cap_style = FALLBACK_POINT_CAPSTYLE
            join_style = FALLBACK_POINT_JOINSTYLE
        elif isinstance(geom, (shapely.LineString, shapely.MultiLineString)):
            cap_style = FALLBACK_LINESTRING_CAPSTYLE

    res = geom.buffer(
        distance=distance, cap_style=cap_style, join_style=join_style, **kwargs
    )

    if res.is_empty:
        return geom

    return res


@passes_kws_to(morphology_buffer)
def erosion(
    geom: BaseGeometry, distance: float = DEFAULT_DISTANCE, **kwargs
) -> BaseGeometry:
    """

    :param distance:
    :param cap_style:
    :param join_style:
    :param geom: The geometry to be eroded
    :return: The eroded geometry
    """
    return morphology_buffer(geom=geom, distance=-distance, **kwargs)


@passes_kws_to(morphology_buffer)
def dilation(
    geom: BaseGeometry, distance: float = DEFAULT_DISTANCE, **kwargs
) -> BaseGeometry:
    """

    :param cap_style:
    :param join_style:
    :param geom: The geometry to be dilated
    :param distance: Dilation amount
    :return: The dilated geometry
    """

    return morphology_buffer(geom=geom, distance=distance, **kwargs)


@passes_kws_to(shapely.geometry.base.BaseGeometry.buffer)
def closing(geom: BaseGeometry, **kwargs) -> BaseGeometry:
    """

    :param geom: The geometry to be closed
    :return: The closed geometry
    """
    return erode(dilate(geom, **kwargs), **kwargs)


@passes_kws_to(shapely.geometry.base.BaseGeometry.buffer)
def opening(geom: BaseGeometry, **kwargs) -> BaseGeometry:
    """

    :param geom: The geometry to be opened
    :return: The opened geometry
    """
    return dilate(erode(geom, **kwargs), **kwargs)


@passes_kws_to(shapely.geometry.base.BaseGeometry.buffer)
def pro_closing(geom: BaseGeometry, **kwargs) -> BaseGeometry:
    """
      Remove Salt and Pepper

      Common Variants
    Opening and closing are themselves often used in combination to achieve more subtle results. If we
    represent the closing of an image f by C(f), and its opening by O(f), then some common combinations include:

    Proper Opening
    Min(f, /em{C}(O(C(f))))

    Proper Closing
    Max(f, O(C(O(f))))

    Automedian Filter
    Max(O(C(O(f))), Min(f, C(O(C(f)))))

    These operators are commonly known as morphological filters.


    Closing
    Dilation means that the central pixel will be replaced by the brightest pixel in the vicinity (filter
    structural element).
    Perfect for removing pepper noise and ensuring that the key features are relatively sharp.

    Opening
    Erosion means is that if we have a structuring element that is a 3 X 3 matrix, the central pixel will be
    replaced by the darkest pixel in the 3 X 3 neighborhood.
    Opening is erosion followed by dilation which makes it perfect for removing salt noise (white dots) and
    ensuring that the key features are relatively sharp.

      :param geom:
      :param kwargs:
      :return:
    """

    return opening(closing(opening(geom, **kwargs), **kwargs), **kwargs)


@passes_kws_to(shapely.geometry.base.BaseGeometry.buffer)
def pro_opening(geom: BaseGeometry, **kwargs) -> BaseGeometry:
    """
      Remove Salt and Pepper

      Common Variants
    Opening and closing are themselves often used in combination to achieve more subtle results. If we
    represent the closing of an image f by C(f), and its opening by O(f), then some common combinations include:

    Proper Opening
    Min(f, /em{C}(O(C(f))))

    Proper Closing
    Max(f, O(C(O(f))))

    Automedian Filter
    Max(O(C(O(f))), Min(f, C(O(C(f)))))

    These operators are commonly known as morphological filters.


    Closing
    Dilation means that the central pixel will be replaced by the brightest pixel in the vicinity (filter
    structural element).
    Perfect for removing pepper noise and ensuring that the key features are relatively sharp.

    Opening
    Erosion means is that if we have a structuring element that is a 3 X 3 matrix, the central pixel will be
    replaced by the darkest pixel in the 3 X 3 neighborhood.
    Opening is erosion followed by dilation which makes it perfect for removing salt noise (white dots) and
    ensuring that the key features are relatively sharp.

      :param geom:
      :param kwargs:
      :return:
    """

    return closing(opening(closing(geom, **kwargs), **kwargs), **kwargs)


# open = opening # keyword clash
erode = erosion
dilate = dilation
close = closing


def zero_buffer(
    geom: BaseGeometry,
) -> BaseGeometry:
    return dilate(geom, distance=0)


class BecameEmptyException(Exception): ...


def clean_shape(
    shape: shapely.geometry.base.BaseGeometry,
    grid_size: Optional[float] = None,
    raise_on_becoming_empty: bool = False,
) -> shapely.geometry.base.BaseGeometry:
    """
    removes self-intersections and duplicate points

    :param raise_on_becoming_empty:
    :param grid_size:
    :param shape: The shape to cleaned
    :return: the cleaned shape
    """

    original_shape = shape

    if isinstance(shape, shapely.Polygon):
        shape = ensure_cw_poly(shape)

    if grid_size is not None:
        if not shape.is_valid:
            try:
                shape = make_valid(shape)
            except shapely.errors.GEOSException as e:
                logger.error(e)

        shape = shapely.set_precision(
            shape,
            grid_size,
            mode="keep_collapsed",
        )

    shape = collapse_duplicate_vertices(shape)

    if not shape.is_valid:
        try:
            shape = make_valid(shape)
        except shapely.errors.GEOSException as e:
            logger.error(e)

    if not original_shape.is_empty:
        if shape.is_empty:
            if raise_on_becoming_empty:
                raise BecameEmptyException(
                    f"{original_shape=} was not empty, became {shape=}"
                )
            else:
                shape = original_shape.representative_point()

    if isinstance(shape, shapely.GeometryCollection):
        if len(shape.geoms) == 1:
            shape = shape.geoms[0]

    return shape


def collapse_duplicate_vertices(
    shape: shapely.geometry.base.BaseGeometry,
) -> shapely.geometry.base.BaseGeometry:
    return zero_buffer(shape).simplify(0)



__all__ = ["read_geometries"]

from pathlib import Path
from typing import Union, Tuple, Any


def read_geometries(
    fn: Union[str, Path], bbox: Tuple[float, float, float, float] = None
) -> Any:
    """
    reads to shapely geometries to features using fiona collection
    feature = dict('geometry': <shapely geometry>, 'properties': <dict with properties>

    :param fn:
    :param bbox:
    :return:
    """
    from fiona import collection
    from shapely.geometry import shape

    with collection(str(fn), "r") as c:
        ft_list = []
        c = c.items(bbox=bbox)
        for ft in c:
            if ft[1]["geometry"] is not None:
                ft_list.append(shape(ft[1]["geometry"]))

    return ft_list



from typing import Optional, Union

import geopandas
import momepy
import shapely

__all__ = ["construct_centerline"]


def construct_centerline(
    input_geometry: shapely.Polygon,
    interpolation_distance: Optional[float] = None,
    truncate_endings: bool = True,
    merge_lines: bool = True,
    simplify_lines: bool = False,
) -> Union[shapely.LineString, shapely.MultiLineString]:
    if interpolation_distance is None:
        interpolation_distance = input_geometry.minimum_clearance

    if interpolation_distance < 1e-15:
        if isinstance(input_geometry, shapely.MultiPolygon):
            return shapely.MultiLineString([f.exterior for f in input_geometry.geoms])
        return input_geometry.exterior

    densified_border = input_geometry.segmentize(interpolation_distance * 0.9)

    voronoi_polys = shapely.voronoi_polygons(
        densified_border,
        only_edges=True,
    )  # equivalent to the scipy.spatial.Voronoi

    center_lines = geopandas.GeoDataFrame(
        geometry=geopandas.GeoSeries(voronoi_polys.geoms)
    ).sjoin(
        geopandas.GeoDataFrame(geometry=geopandas.GeoSeries(input_geometry)),
        predicate="within",
    )  # to select only the linestring within the input geometry

    if truncate_endings:
        graph = momepy.gdf_to_nx(center_lines)

        graph.remove_nodes_from(
            node for node, degree in dict(graph.degree()).items() if degree < 2
        )

        center_lines = momepy.nx_to_gdf(graph, points=False)

    ret = center_lines.unary_union

    if merge_lines:
        ret = shapely.line_merge(ret)

    if simplify_lines:
        ret = shapely.simplify(ret, tolerance=interpolation_distance * 0.1)

    if isinstance(ret, shapely.geometry.MultiLineString):
        geoms = list(ret.geoms)
        if len(geoms) == 1:
            return geoms[0]

    return ret



__author__ = "Christian Heider Lindbjerg"
__doc__ = r"""

           Created on 02-12-2020
           """

import time
from typing import Any

# noinspection PyUnresolvedReferences
from qgis.PyQt import QtGui

# noinspection PyUnresolvedReferences
from qgis.PyQt.QtCore import Qt

__all__ = ["qt_draw_timestamp"]

from warg import Triple


def qt_draw_timestamp(
    image: Any,  # QtImage
    font_size: int = 10,
    font_color: Triple = (255, 255, 255),
    font_family: str = "Arial",
    font_style: int = QtGui.QFont.StyleNormal,
    font_weight: int = QtGui.QFont.Normal,
) -> None:
    """

    :param image:
    :param font_size:
    :param font_color:
    :param font_family:
    :param font_style:
    :param font_weight:
    :return:
    """
    timestamp = time.strftime("%Y-%m-%d %H:%M:%S")
    font = QtGui.QFont(font_family, font_size, font_style, font_weight)
    font.setPixelSize(font_size)
    font.setBold(True)
    painter = QtGui.QPainter(image)
    painter.setFont(font)
    painter.setPen(QtGui.QColor(*font_color))
    painter.drawText(
        0,
        0,
        image.width(),
        image.height(),
        Qt.AlignCenter,
        timestamp,
    )
    painter.end()



__author__ = "heider"
__doc__ = r"""

           Created on 1/23/23
           """

from pathlib import Path

with open(Path(__file__).parent / "README.md") as this_init_file:
    __doc__ += this_init_file.read()

from .well_known_binary import *
from .well_known_text import *



import logging
from typing import Any, Iterable, Sequence

# noinspection PyUnresolvedReferences
from qgis.core import QgsDefaultValue, QgsEditorWidgetSetup, QgsFieldConstraints

__all__ = [
    "add_dropdown_widget",
    "make_field_unique",
    "HIDDEN_WIDGET",
    "make_field_not_null",
    "make_field_default",
    "make_field_boolean",
    "make_field_reuse_last_entered_value",
]

IGNORE_THIS_STRING = """

Line edit – a simple edit box
Classification – displays a combo box with the values used for “unique value” classification (symbology tab)
Range – allows numeric values within a given range, the widget can be either slider or spin box
Unique values
    editable – displays a line edit widget with auto-completion suggesting values already used in the
    attribute table
    not editable – displays a combo box with already used values
File name – adds a file chooser dialog
Value map – shows a combo box with predefined description/value items
Enumeration – a combo box with values that can be used within the columns type
Immutable – read-only
Hidden – makes the attribute invisible for the user
CheckBox – a checkbox with customizable representation for both checked and unchecked state
Text edit – an edit box that allow multiple input lines
Calendar – a calendar widget to input dates
"""

IGNORE_THIS_STRING2 = """
QGIS Widget Types
_________________

Binary

Checkbox

Classification

Color

DateTime

Enumeration

Attachment

Geometry

Hidden

JsonView

KeyValue

List

Range

RelationReference

TextEdit

UniqueValues

UuidGenerator

ValueMap

ValueRelation

"""

HIDDEN_WIDGET = QgsEditorWidgetSetup("Hidden", {})
CHECKBOX_WIDGET = QgsEditorWidgetSetup(
    "CheckBox",
    {
        "Checked state": "True",
        "Unchecked state": "False",
    },
)
UNIQUE_VALUES_WIDGET = QgsEditorWidgetSetup(
    "UniqueValues",
    {"Editable": True},
)

logger = logging.getLogger(__name__)


def add_dropdown_widget(layers: Any, field_name: str, form_widget: Any) -> None:
    """
    https://gis.stackexchange.com/questions/470963/setting-dropdown-on-feature-attribute-form-using-plugin
      :param layer:
      :param field_name:
      :param form_widget:
      :return:
    """

    if layers is None:
        return

    for layers_inner in layers:
        if layers_inner:
            if isinstance(layers_inner, Iterable):
                for layer in layers_inner:
                    if layer:
                        idx = layer.fields().indexFromName(field_name)
                        if idx < 0:
                            continue
                        layer.setEditorWidgetSetup(
                            idx,
                            form_widget,
                        )
            else:
                idx = layers_inner.fields().indexFromName(field_name)
                if idx < 0:
                    continue
                layers_inner.setEditorWidgetSetup(
                    idx,
                    form_widget,
                )


def make_field_unique(
    layers: Sequence[Any], *, field_name: str = "id", auto_generate: bool = True
) -> None:
    if layers is None:
        return

    unique_widget = None
    default_value_generator = None

    if False:
        unique_widget = QgsEditorWidgetSetup(
            "UuidGenerator",
            {},
        )
        if False:
            logger.error(unique_widget.config())
    elif auto_generate:
        default_value_generator = QgsDefaultValue()
        default_value_generator.setExpression("rtrim( ltrim( uuid(), '{'), '}')")
    else:
        unique_widget = UNIQUE_VALUES_WIDGET

    for layers_inner in layers:
        if layers_inner:
            if isinstance(layers_inner, Iterable):
                for layer in layers_inner:
                    if layer:
                        idx = layer.fields().indexFromName(field_name)
                        if idx < 0:
                            continue

                        if unique_widget:
                            layer.setEditorWidgetSetup(
                                idx,
                                unique_widget,
                            )
                        elif default_value_generator:
                            layer.setDefaultValueDefinition(
                                idx, default_value_generator
                            )
                        else:
                            raise NotImplementedError

                        layer.setFieldConstraint(
                            idx, QgsFieldConstraints.ConstraintNotNull
                        )
                        layer.setFieldConstraint(
                            idx, QgsFieldConstraints.ConstraintUnique
                        )
            else:
                idx = layers_inner.fields().indexFromName(field_name)

                if idx < 0:
                    continue

                if unique_widget:
                    layers_inner.setEditorWidgetSetup(
                        idx,
                        unique_widget,
                    )
                elif default_value_generator:
                    layers_inner.setDefaultValueDefinition(idx, default_value_generator)
                else:
                    raise NotImplementedError

                layers_inner.setFieldConstraint(
                    idx, QgsFieldConstraints.ConstraintNotNull
                )
                layers_inner.setFieldConstraint(
                    idx, QgsFieldConstraints.ConstraintUnique
                )


def make_field_not_null(layers: Sequence[Any], field_name: str = "name") -> None:
    if layers is None:
        return

    for layers_inner in layers:
        if layers_inner:
            if isinstance(layers_inner, Iterable):
                for layers in layers_inner:
                    if layers:
                        idx = layers.fields().indexFromName(field_name)
                        if idx < 0:
                            continue

                        layers.setFieldConstraint(
                            idx, QgsFieldConstraints.ConstraintNotNull
                        )
            else:
                idx = layers_inner.fields().indexFromName(field_name)
                if idx < 0:
                    continue
                layers_inner.setFieldConstraint(
                    idx, QgsFieldConstraints.ConstraintNotNull
                )


def make_field_default(
    layers: Sequence[Any], field_name: str, default_expression: str = "'None'"
) -> None:
    if layers is None:
        return

    default_value = QgsDefaultValue()
    default_value.setExpression(default_expression)

    for layers_inner in layers:
        if layers_inner:
            if isinstance(layers_inner, Iterable):
                for layers in layers_inner:
                    if layers:
                        layers.setDefaultValueDefinition(
                            layers.fields().indexFromName(field_name), default_value
                        )
            else:
                layers_inner.setDefaultValueDefinition(
                    layers_inner.fields().indexFromName(field_name), default_value
                )


def make_field_boolean(layers: Sequence[Any], field_name: str) -> None:
    if layers is None:
        return

    for layers_inner in layers:
        if layers_inner:
            if isinstance(layers_inner, Iterable):
                for layers in layers_inner:
                    if layers:
                        idx = layers.fields().indexFromName(field_name)

                        if idx < 0:
                            continue

                        layers.setDefaultValueDefinition(idx, CHECKBOX_WIDGET)
            else:
                idx = layers_inner.fields().indexFromName(field_name)

                if idx < 0:
                    continue

                layers_inner.setEditorWidgetSetup(
                    idx,
                    CHECKBOX_WIDGET,
                )


def make_field_reuse_last_entered_value(layers: Sequence[Any], field_name: str) -> None:
    if layers is None:
        return

    for layers_inner in layers:
        if layers_inner:
            if isinstance(layers_inner, Iterable):
                for layer in layers_inner:
                    if layer:
                        idx = layer.fields().indexFromName(field_name)

                        if idx < 0:
                            continue

                        layer_form_config = layer.editFormConfig()
                        layer_form_config.setReuseLastValue(idx, True)
                        layer.setEditFormConfig(layer_form_config)
            else:
                idx = layers_inner.fields().indexFromName(field_name)

                if idx < 0:
                    continue

                layer_form_config = layers_inner.editFormConfig()
                layer_form_config.setReuseLastValue(idx, True)
                layers_inner.setEditFormConfig(layer_form_config)


def fit_field_to_length(layers: Sequence[Any], field_name: str, length: int) -> None:
    if layers is None:
        return

    for layers_inner in layers:
        if layers_inner:
            if isinstance(layers_inner, Iterable):
                for layer in layers_inner:
                    if layer:
                        fields = layer.fields()

                        idx = fields.indexFromName(field_name)
                        if idx < 0:
                            continue

                        fields[idx].setLength(length)
            else:
                fields = layers_inner.fields()
                idx = fields.indexFromName(field_name)
                if idx < 0:
                    continue
                fields[idx].setLength(length)



from enum import Enum

from jord.gdal_utilities import GDAL

__all__ = ["gdal_error_type_map", "GdalErrorTypeEnum", "gdal_error_handler"]


class GdalErrorTypeEnum(Enum):
    none = GDAL.CE_None
    debug = GDAL.CE_Debug
    warning = GDAL.CE_Warning
    failure = GDAL.CE_Failure
    fatal = GDAL.CE_Fatal


gdal_error_type_map = {v.value: str(v.name).capitalize() for v in GdalErrorTypeEnum}


def gdal_error_handler(err_class, err_num, err_msg) -> None:
    err_msg = err_msg.replace("\n", " ")
    err_class = gdal_error_type_map.get(err_class, "None")
    print(f"Error Number: {err_num}")
    print(f"Error Type: {err_class}")
    print(f"Error Message: {err_msg}")


if __name__ == "__main__":
    print(gdal_error_type_map)
    ...
    # GDAL.PushErrorHandler(gdal_error_handler)



__author__ = "heider"
__doc__ = r"""

           Created on 1/16/23
           """

__all__ = ["SilenceGDALSession"]

from warg import AlsoDecorator

from .importing import GDAL


class SilenceGDALSession(AlsoDecorator):
    """
    Session for silencing gdal warning and errors.
    TODO: add support for having a lasting side effect or leaving last set state of error/exception handling
    """

    def __init__(self): ...

    def __enter__(self) -> bool:
        GDAL.PushErrorHandler(
            "CPLQuietErrorHandler"
        )  # Stop GDAL printing both warnings and errors to STDERR
        GDAL.UseExceptions()  # Make GDAL raise python exceptions for errors (warnings won't raise an exception)
        return True

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        # gdal.GetUseExceptions()
        GDAL.PopErrorHandler()  # Pop error handler previously pushed
        GDAL.DontUseExceptions()  # Dont use exception anymore
        # gdal.ErrorReset()


if __name__ == "__main__":
    with SilenceGDALSession():
        ...



import shapely

from shapely import LinearRing
from shapely.geometry import LineString, MultiLineString, Point, Polygon
from shapely.ops import linemerge

__all__ = ["ensure_ccw_ring", "ensure_cw_ring", "split_ring"]

from typing import Union


def ensure_ccw_ring(ring: LinearRing) -> LinearRing:
    if not ring.is_ccw:
        return LinearRing(list(ring.coords)[::-1])
    return ring


def ensure_cw_ring(ring: LinearRing) -> LinearRing:
    if ring.is_ccw:
        return LinearRing(list(ring.coords)[::-1])
    return ring


def split_ring(
    ring: LinearRing,
    split: Union[LinearRing, LineString, MultiLineString, shapely.GeometryCollection],
):
    """Split a linear ring geometry, returns a [Multi]LineString

    See my PostGIS function on scigen named ST_SplitRing
    """
    valid_types = ("MultiLineString", "LineString", "GeometryCollection")
    if not hasattr(ring, "geom_type"):
        raise ValueError("expected ring as a geometry")
    elif not hasattr(split, "geom_type"):
        raise ValueError("expected split as a geometry")
    if ring.geom_type == "LinearRing":
        ring = LineString(ring)
    if ring.geom_type != "LineString":
        raise ValueError(
            "ring is not a LinearRing or LineString, found " + str(ring.geom_type)
        )
    elif not ring.is_closed:
        raise ValueError("ring is not closed")
    elif split.is_empty:
        return ring
    elif not split.intersects(ring):
        # split does not intersect ring
        return ring
    if split.geom_type == "LinearRing":
        split = LineString(split)
    if split.geom_type not in valid_types:
        raise ValueError(
            "split is not a LineString-like or GeometryCollection geometry, "
            "found " + str(split.geom_type)
        )

    intersections = ring.intersection(split)
    if intersections.is_empty:
        # no intersections, returning same ring
        return ring
    elif intersections.geom_type == "Point":
        # Simple case, where there is only one line intersecting the ring
        result = Polygon(ring).difference(split).exterior
        # If it is a coordinate of the ring, then the ring needs to be rotated
        coords = result.coords[:-1]
        found_i = 0
        for i, c in enumerate(coords):
            if Point(c).almost_equals(intersections):
                found_i = i
                break
        if found_i > 0:
            result = Polygon(coords[i:] + coords[:i]).exterior
        if result.interpolate(0).distance(intersections) > 0:
            raise Exception(
                "result start point %s to intersection %s is %s"
                % (result.interpolate(0), intersections, result.distance(intersections))
            )
        elif result.geom_type != "LinearRing":
            raise Exception("result is not a LinearRing, found " + result.geom_type)
        elif not result.is_closed:
            raise Exception("result is not closed")
        return LineString(result)

    difference = ring.difference(split)
    if difference.geom_type != "MultiLineString":
        raise ValueError(
            "expected MultiLineString difference, found " + difference.geom_type
        )

    start_point = ring.interpolate(0)
    if start_point.distance(intersections) == 0:
        # special case: start point is the same as an intersection
        return difference

    # Otherwise the line where the close meets needs to be fused
    fuse = []
    parts = list(difference.geoms)
    for ipart, part in enumerate(parts):
        if part.intersects(start_point):
            fuse.append(ipart)
    if len(fuse) != 2:
        raise ValueError("expected 2 geometries, found " + str(len(fuse)))
    # glue the last to the first
    popped_part = parts.pop(fuse[1])
    parts[fuse[0]] = linemerge([parts[fuse[0]], popped_part])
    return MultiLineString(parts)



__author__ = "heider"
__doc__ = r"""

           Created on 5/5/22
           """

from pathlib import Path

with open(Path(__file__).parent / "README.md") as this_init_file:
    __doc__ += this_init_file.read()

from .clients import *
from .client import *
from .layer_creation import *
from .layer_serialisation import *
from .pandas_procedures import *
from .parsing import *
from .procedures import *
from .serialisation import *
from .type_solving import *
from .uri_utilities import *



from enum import Enum

__all__ = ["CommonVectorDataProviderLibEnum"]


class CommonVectorDataProviderLibEnum(Enum):
    memory = "memory"
    postgres = "postgres"


if __name__ == "__main__":

    def auishd(iface) -> None:
        """
        Some services require you to supply your own API key for the services to work.

        For Esri basemaps you will need a valid ArcGIS online subscription to use the maps.
        """
        from PyQt5.QtCore import QSettings

        """
connection_type
connection_name
authcfg
password
referer
url
username
zmax
zmin
"""

        # Sources
        sources = []
        sources.append(
            [
                "connections-xyz",
                "Google Maps",
                "",
                "",
                "",
                "https://mt1.google.com/vt/lyrs=m&x=%7Bx%7D&y=%7By%7D&z=%7Bz%7D",
                "",
                "19",
                "0",
            ]
        )
        sources.append(
            [
                "connections-xyz",
                "Google Satellite",
                "",
                "",
                "",
                "https://mt1.google.com/vt/lyrs=s&x=%7Bx%7D&y=%7By%7D&z=%7Bz%7D",
                "",
                "19",
                "0",
            ]
        )
        sources.append(
            [
                "connections-xyz",
                "Google Terrain",
                "",
                "",
                "",
                "https://mt1.google.com/vt/lyrs=t&x=%7Bx%7D&y=%7By%7D&z=%7Bz%7D",
                "",
                "19",
                "0",
            ]
        )
        sources.append(
            [
                "connections-xyz",
                "Google Terrain Hybrid",
                "",
                "",
                "",
                "https://mt1.google.com/vt/lyrs=p&x=%7Bx%7D&y=%7By%7D&z=%7Bz%7D",
                "",
                "19",
                "0",
            ]
        )
        sources.append(
            [
                "connections-xyz",
                "Google Satellite Hybrid",
                "",
                "",
                "",
                "https://mt1.google.com/vt/lyrs=y&x=%7Bx%7D&y=%7By%7D&z=%7Bz%7D",
                "",
                "19",
                "0",
            ]
        )
        sources.append(
            [
                "connections-xyz",
                "Stamen Terrain",
                "",
                "",
                "Map tiles by Stamen Design, under CC BY 3.0. Data by OpenStreetMap, under ODbL",
                "http://tile.stamen.com/terrain/%7Bz%7D/%7Bx%7D/%7By%7D.png",
                "",
                "20",
                "0",
            ]
        )
        sources.append(
            [
                "connections-xyz",
                "Stamen Toner",
                "",
                "",
                "Map tiles by Stamen Design, under CC BY 3.0. Data by OpenStreetMap, under ODbL",
                "http://tile.stamen.com/toner/%7Bz%7D/%7Bx%7D/%7By%7D.png",
                "",
                "20",
                "0",
            ]
        )
        sources.append(
            [
                "connections-xyz",
                "Stamen Toner Light",
                "",
                "",
                "Map tiles by Stamen Design, under CC BY 3.0. Data by OpenStreetMap, under ODbL",
                "http://tile.stamen.com/toner-lite/%7Bz%7D/%7Bx%7D/%7By%7D.png",
                "",
                "20",
                "0",
            ]
        )
        sources.append(
            [
                "connections-xyz",
                "Stamen Watercolor",
                "",
                "",
                "Map tiles by Stamen Design, under CC BY 3.0. Data by OpenStreetMap, under ODbL",
                "http://tile.stamen.com/watercolor/%7Bz%7D/%7Bx%7D/%7By%7D.jpg",
                "",
                "18",
                "0",
            ]
        )
        sources.append(
            [
                "connections-xyz",
                "Wikimedia Map",
                "",
                "",
                "OpenStreetMap contributors, under ODbL",
                "https://maps.wikimedia.org/osm-intl/%7Bz%7D/%7Bx%7D/%7By%7D.png",
                "",
                "20",
                "1",
            ]
        )
        sources.append(
            [
                "connections-xyz",
                "Wikimedia Hike Bike Map",
                "",
                "",
                "OpenStreetMap contributors, under ODbL",
                "http://tiles.wmflabs.org/hikebike/%7Bz%7D/%7Bx%7D/%7By%7D.png",
                "",
                "17",
                "1",
            ]
        )
        sources.append(
            [
                "connections-xyz",
                "Esri Boundaries Places",
                "",
                "",
                "Requires ArcGIS Onlinesubscription",
                "https://server.arcgisonline.com/ArcGIS/rest/services/Reference/World_Boundaries_and_Places"
                "/MapServer/tile/%7Bz%7D/%7By%7D/%7Bx%7D",
                "",
                "20",
                "0",
            ]
        )
        sources.append(
            [
                "connections-xyz",
                "Esri Gray (dark)",
                "",
                "",
                "Requires ArcGIS Onlinesubscription",
                "http://services.arcgisonline.com/ArcGIS/rest/services/Canvas/World_Dark_Gray_Base/MapServer/tile/%7Bz%7D/%7By%7D/%7Bx%7D",
                "",
                "16",
                "0",
            ]
        )
        sources.append(
            [
                "connections-xyz",
                "Esri Gray (light)",
                "",
                "",
                "Requires ArcGIS Onlinesubscription",
                "http://services.arcgisonline.com/ArcGIS/rest/services/Canvas/World_Light_Gray_Base/MapServer/tile/%7Bz%7D/%7By%7D/%7Bx%7D",
                "",
                "16",
                "0",
            ]
        )
        sources.append(
            [
                "connections-xyz",
                "Esri National Geographic",
                "",
                "",
                "Requires ArcGIS Onlinesubscription",
                "http://services.arcgisonline.com/ArcGIS/rest/services/NatGeo_World_Map/MapServer/tile/%7Bz%7D/%7By%7D/%7Bx%7D",
                "",
                "12",
                "0",
            ]
        )
        sources.append(
            [
                "connections-xyz",
                "Esri Ocean",
                "",
                "",
                "Requires ArcGIS Onlinesubscription",
                "https://services.arcgisonline.com/ArcGIS/rest/services/Ocean/World_Ocean_Base/MapServer/tile/%7Bz%7D/%7By%7D/%7Bx%7D",
                "",
                "10",
                "0",
            ]
        )
        sources.append(
            [
                "connections-xyz",
                "Esri Satellite",
                "",
                "",
                "Requires ArcGIS Onlinesubscription",
                "https://server.arcgisonline.com/ArcGIS/rest/services/World_Imagery/MapServer/tile/%7Bz%7D/%7By%7D/%7Bx%7D",
                "",
                "17",
                "0",
            ]
        )
        sources.append(
            [
                "connections-xyz",
                "Esri Standard",
                "",
                "",
                "Requires ArcGIS Onlinesubscription",
                "https://server.arcgisonline.com/ArcGIS/rest/services/World_Street_Map/MapServer/tile/%7Bz%7D/%7By%7D/%7Bx%7D",
                "",
                "17",
                "0",
            ]
        )
        sources.append(
            [
                "connections-xyz",
                "Esri Terrain",
                "",
                "",
                "Requires ArcGIS Onlinesubscription",
                "https://server.arcgisonline.com/ArcGIS/rest/services/World_Terrain_Base/MapServer/tile/%7Bz%7D/%7By%7D/%7Bx%7D",
                "",
                "13",
                "0",
            ]
        )
        sources.append(
            [
                "connections-xyz",
                "Esri Transportation",
                "",
                "",
                "Requires ArcGIS Onlinesubscription",
                "https://server.arcgisonline.com/ArcGIS/rest/services/Reference/World_Transportation/MapServer/tile/%7Bz%7D/%7By%7D/%7Bx%7D",
                "",
                "20",
                "0",
            ]
        )
        sources.append(
            [
                "connections-xyz",
                "Esri Topo World",
                "",
                "",
                "Requires ArcGIS Onlinesubscription",
                "http://services.arcgisonline.com/ArcGIS/rest/services/World_Topo_Map/MapServer/tile/%7Bz%7D/%7By%7D/%7Bx%7D",
                "",
                "20",
                "0",
            ]
        )
        sources.append(
            [
                "connections-xyz",
                "OpenStreetMap Standard",
                "",
                "",
                "OpenStreetMap contributors, under ODbL",
                "http://tile.openstreetmap.org/%7Bz%7D/%7Bx%7D/%7By%7D.png",
                "",
                "19",
                "0",
            ]
        )
        sources.append(
            [
                "connections-xyz",
                "OpenStreetMap H.O.T.",
                "",
                "",
                "OpenStreetMap contributors, under ODbL",
                "http://tile.openstreetmap.fr/hot/%7Bz%7D/%7Bx%7D/%7By%7D.png",
                "",
                "19",
                "0",
            ]
        )
        sources.append(
            [
                "connections-xyz",
                "OpenStreetMap Monochrome",
                "",
                "",
                "OpenStreetMap contributors, under ODbL",
                "http://tiles.wmflabs.org/bw-mapnik/%7Bz%7D/%7Bx%7D/%7By%7D.png",
                "",
                "19",
                "0",
            ]
        )
        sources.append(
            [
                "connections-xyz",
                "OpenTopoMap",
                "",
                "",
                "Kartendaten: © OpenStreetMap-Mitwirkende, SRTM | Kartendarstellung: © OpenTopoMap (CC-BY-SA)",
                "https://tile.opentopomap.org/%7Bz%7D/%7Bx%7D/%7By%7D.png",
                "",
                "17",
                "1",
            ]
        )
        sources.append(
            [
                "connections-xyz",
                "Strava All",
                "",
                "",
                "OpenStreetMap contributors, under ODbL",
                "https://heatmap-external-b.strava.com/tiles/all/bluered/%7Bz%7D/%7Bx%7D/%7By%7D.png",
                "",
                "15",
                "0",
            ]
        )
        sources.append(
            [
                "connections-xyz",
                "Strava Run",
                "",
                "",
                "OpenStreetMap contributors, under ODbL",
                "https://heatmap-external-b.strava.com/tiles/run/bluered/%7Bz%7D/%7Bx%7D/%7By%7D.png?v=19",
                "",
                "15",
                "0",
            ]
        )
        sources.append(
            [
                "connections-xyz",
                "Open Weather Map Temperature",
                "",
                "",
                "Map tiles by OpenWeatherMap, under CC BY-SA 4.0",
                "http://tile.openweathermap.org/map/temp_new/%7Bz%7D/%7Bx%7D/%7By%7D.png?APPID={your_API_key}",
                "",
                "19",
                "0",
            ]
        )
        sources.append(
            [
                "connections-xyz",
                "Open Weather Map Clouds",
                "",
                "",
                "Map tiles by OpenWeatherMap, under CC BY-SA 4.0",
                "http://tile.openweathermap.org/map/clouds_new/%7Bz%7D/%7Bx%7D/%7By%7D.png?APPID={your_API_key}",
                "",
                "19",
                "0",
            ]
        )
        sources.append(
            [
                "connections-xyz",
                "Open Weather Map Wind Speed",
                "",
                "",
                "Map tiles by OpenWeatherMap, under CC BY-SA 4.0",
                "http://tile.openweathermap.org/map/wind_new/%7Bz%7D/%7Bx%7D/%7By%7D.png?APPID={your_API_key}",
                "",
                "19",
                "0",
            ]
        )
        sources.append(
            [
                "connections-xyz",
                "CartoDb Dark Matter",
                "",
                "",
                "Map tiles by CartoDB, under CC BY 3.0. Data by OpenStreetMap, under ODbL.",
                "http://basemaps.cartocdn.com/dark_all/%7Bz%7D/%7Bx%7D/%7By%7D.png",
                "",
                "20",
                "0",
            ]
        )
        sources.append(
            [
                "connections-xyz",
                "CartoDb Positron",
                "",
                "",
                "Map tiles by CartoDB, under CC BY 3.0. Data by OpenStreetMap, under ODbL.",
                "http://basemaps.cartocdn.com/light_all/%7Bz%7D/%7Bx%7D/%7By%7D.png",
                "",
                "20",
                "0",
            ]
        )
        sources.append(
            [
                "connections-xyz",
                "Bing VirtualEarth",
                "",
                "",
                "",
                "http://ecn.t3.tiles.virtualearth.net/tiles/a{q}.jpeg?g=1",
                "",
                "19",
                "1",
            ]
        )

        for source in sources:  # Add sources to browser
            connection_type = source[0]
            connection_name = source[1]

            QSettings().setValue(
                f"qgis/{connection_type}/{connection_name}/authcfg", source[2]
            )
            QSettings().setValue(
                f"qgis/{connection_type}/{connection_name}/password", source[3]
            )
            QSettings().setValue(
                f"qgis/{connection_type}/{connection_name}/referer", source[4]
            )
            QSettings().setValue(
                f"qgis/{connection_type}/{connection_name}/url", source[5]
            )
            QSettings().setValue(
                f"qgis/{connection_type}/{connection_name}/username", source[6]
            )
            QSettings().setValue(
                f"qgis/{connection_type}/{connection_name}/zmax", source[7]
            )
            QSettings().setValue(
                f"qgis/{connection_type}/{connection_name}/zmin", source[8]
            )

        iface.reloadConnections()  # Update GUI
